from isr_matcher._constants._filters import ISR_FILTERS
from isr_matcher._constants._parse import INT_TYPES, FLOAT_TYPES
from isr_matcher.geometry.track_segment import TrackSegment
from isr_matcher.geometry.operational_point import OperationalPoint
from isr_matcher.data_handlers.transformer import Transformer
from isr_matcher.ops._preprocessing import ISRPreprocessor
from isr_matcher._constants._parse import extract_station_names
from isr_matcher._constants._properties import (
    ISR_PROPERTIES_OPERATIONAL_POINT,
    ISR_PROPERTIES_JUNCTION,
    ISR_PROPERTIES_TRACK_SEGMENTS,
)
from isr_matcher._constants._filters import ISR_EXCEPTIONAL_STATION_NAMES
from isr_matcher._constants.logging import setup_logger
from pathlib import Path
from requests import post
from requests.exceptions import RequestException
import json
from json import dump as json_dump, loads as json_load
from typing import Literal, Union, Tuple, TypeVar, overload
from shapely.geometry.base import BaseGeometry
from shapely import Point, LineString, MultiLineString
import numpy as np
import pickle
from copy import deepcopy
import logging

# Type T: only instances of class BaseGeometry or subclasses (Point, Linestring, etc.)
T = TypeVar("T", bound="BaseGeometry")

# Create logger for the current module
setup_logger()
logger = logging.getLogger(__name__)


class RailDataImport:

    """
    Class that manages the aquisition of railway infrastructure data from ISR.

    Attributes
    ----------
    _filters : list
        A list of filters used for data acquisition.
    _project_path : Path
        The path to the project directory.
    _special_chars : list
        A list of special characters used in the data.
    """

    def __init__(self):
        self._filters = [
            "EQUALS_TRACK",
            "EQUALS_OP",
            "EQUALS_TRANSITION",
            "EQUALS_TRACK_SEG",
            "TRACKS_IN_BBOX",
            "OP_IN_BBOX",
            "TRANSITIONS_IN_BBOX",
            "TUNNEL_IN_BBOX",
            "BBOX",
            "TRACKS_IN_BOUNDARY",
            "OP_IN_BOUNDARY",
            "TRANSITIONS_IN_BOUNDARY",
            "TUNNEL_IN_BOUNDARY",
            "BOUNDARY",
        ]
        self._project_path = Path(__file__).parent.parent.parent.parent
        self._special_chars = ["Ä", "Ö", "Ü", "ä", "ö", "ü", "ß"]

    # define attributes as read-only
    @property
    def filters(self):
        return self._filters

    @property
    def project_path(self):
        return self._project_path

    def query_isr(
        self,
        filter_name: Literal[
            "EQUALS_TRACK",
            "EQUALS_OP",
            "EQUALS_TRANSITION",
            "EQUALS_TRACK_SEG",
            "TRACKS_IN_BBOX",
            "OP_IN_BBOX",
            "TRANSITIONS_IN_BBOX",
            "TUNNEL_IN_BBOX",
            "BBOX",
            "TRACKS_IN_BOUNDARY",
            "OP_IN_BOUNDARY",
            "TRANSITIONS_IN_BOUNDARY",
            "TUNNEL_IN_BOUNDARY",
            "BOUNDARY",
        ],
        args: list[str | int | float],
    ) -> dict | None:
        """
        Queries ISR based on the specified filter and arguments.

        Parameters
        ----------
        filter_name: Literal
            The type of filter to be applied. Choose from: TODO: UPDATE
                - 'EQUALS_TRACK': Query by track. 'args' is a tuple of (<track_nr>).
                - 'EQUALS_OP': Query by operational point. 'args' is a tuple of (<name>, <track_nr>).
                - 'EQUALS_TRACK_SEG': Query by track segment. 'args' is a tuple of (<track_nr>, <name>).
                - 'TRACKS_IN_BBOX': Query tracks within a bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'OP_IN_BBOX': Query operational points within a bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'TRANSITIONS_IN_BBOX': Query transitions within a bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'BBOX': Query using a general bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'BOUNDARY': Query using a boundary (LinearRing). 'args' is a tuple of (<string of coordinates seperated by space: "lon_1 lat_1 lon_2 lat_2 ... ">).

        args: list
            A list of arguments corresponding to the chosen filter. The interpretation of arguments depends on the selected filter.

        Returns
        -------
            response: dict
                A dictionary in JSON format if the query is successful; otherwise, returns None.

        Notes
        -----
        - The method validates the input parameters before creating the filter string.
        - Special characters in arguments are replaced by wildcards.
        - The method sends a request to the ISR API and returns the JSON response if the request is successful.
        - If the request fails or encounters an exception, None is returned.

        """

        # check input
        self._check_input_for_query(filter_name=filter_name, args=args)

        # create filter string
        filter_str = ISR_FILTERS[filter_name]
        for i, arg in enumerate(args):
            # replace special chars by wildcard
            new_arg = arg
            for char in self._special_chars:
                if isinstance(new_arg, str) and char in new_arg:
                    new_arg = new_arg.replace(char, '*')
            filter_str = filter_str.replace(f"$arg{i + 1}", str(new_arg))

        # send the request
        try:
            response = post(
                "https://geovdbn.deutschebahn.com/pgv-map/geoserver.action",
                data=filter_str,
                headers={"Content-Type": "application/json"},
            )
        except Exception:
            response = None

        if response is not None and response.status_code == 200:
            # The request was successful
            return response.json()
        else:
            # request was not successful
            return None

    @overload
    def query(
        self,
        filter_name: Literal["EQUALS_TRACK"],
        args: list[str | int | float],
        enhance_kilometrage: bool,
        cache_preprocessed_segments: bool,
    ) -> list[TrackSegment]:
        ...

    @overload
    def query(
        self,
        filter_name: Literal["EQUALS_TRACK_SEG"],
        args: list[str | int | float],
        enhance_kilometrage: bool,
        cache_preprocessed_segments: bool,
    ) -> TrackSegment:
        ...

    @overload
    def query(
        self,
        filter_name: Literal["EQUALS_OP"],
        args: list[str | int | float],
        enhance_kilometrage: bool,
        cache_preprocessed_segments: bool,
    ) -> OperationalPoint:
        ...

    @overload
    def query(
        self,
        filter_name: Literal["TRACKS_IN_BBOX"],
        args: list[str | int | float],
        enhance_kilometrage: bool,
        cache_preprocessed_segments: bool,
    ) -> list[TrackSegment]:
        ...

    @overload
    def query(
        self,
        filter_name: Literal["OP_IN_BBOX"],
        args: list[str | int | float],
        enhance_kilometrage: bool,
        cache_preprocessed_segments: bool,
    ) -> list[OperationalPoint]:
        ...

    @overload
    def query(
        self,
        filter_name: Literal["TRANSITIONS_IN_BBOX"],
        args: list[str | int | float],
        enhance_kilometrage: bool,
        cache_preprocessed_segments: bool,
    ) -> list[OperationalPoint]:
        ...

    @overload
    def query(
        self,
        filter_name: Literal["BBOX"],
        args: list[str | int | float],
        enhance_kilometrage: bool,
        cache_preprocessed_segments: bool,
    ) -> list[OperationalPoint | TrackSegment]:
        ...

    @overload
    def query(
        self,
        filter_name: Literal["BOUNDARY"],
        args: list[str | int | float],
        enhance_kilometrage: bool,
        cache_preprocessed_segments: bool,
    ) -> list[OperationalPoint | TrackSegment]:
        ...

    def query(
        self,
        filter_name: Literal[
            "EQUALS_TRACK",
            "EQUALS_OP",
            "EQUALS_TRANSITION",
            "EQUALS_TRACK_SEG",
            "TRACKS_IN_BBOX",
            "OP_IN_BBOX",
            "TRANSITIONS_IN_BBOX",
            "TUNNEL_IN_BBOX",
            "BBOX",
            "TRACKS_IN_BOUNDARY",
            "OP_IN_BOUNDARY",
            "TRANSITIONS_IN_BOUNDARY",
            "TUNNEL_IN_BOUNDARY",
            "BOUNDARY",
        ],
        args: list[str | int | float],
        enhance_kilometrage: bool = True,
        cache_preprocessed_segments: bool = True,
    ) -> Union[TrackSegment, list[TrackSegment], OperationalPoint, list[OperationalPoint]]:
        """
        Queries for objects according to 'filter_name' and 'args'. Objects are read from cache, if possible. If the object does not exist in the cache, a query is sent to ISR, and the object is created from the response.
        For filters BBOX and BOUNDARY, a query is sent to ISR with the geometric constraint. For each feature in the response, it is checked if the object can be read from the cache or if it must be created from the feature.

        Parameters
        ----------
        filter_name: Literal
            The type of filter to be applied. Choose from: TODO: Update
                - 'EQUALS_TRACK': Query by track. 'args' is a tuple of (<track_nr>).
                - 'EQUALS_OP': Query by operational point. 'args' is a tuple of (<name>, <track_nr>).
                - 'EQUALS_TRACK_SEG': Query by track segment. 'args' is a tuple of (<track_nr>, <name>).
                - 'TRACKS_IN_BBOX': Query tracks within a bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'OP_IN_BBOX': Query operational points within a bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'TRANSITIONS_IN_BBOX': Query transitions within a bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'BBOX': Query using a general bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'BOUNDARY': Query using a boundary (LinearRing). 'args' is a tuple of (<string of coordinates separated by space: "lon_1 lat_1 lon_2 lat_2 ... ">).

        args: list
            A list of arguments corresponding to the chosen filter. The interpretation of arguments depends on the selected filter.

        Returns
        -------
        response: Union[TrackSegment, list[TrackSegment], OperationalPoint, list[OperationalPoint]]
            Returns the queried objects. The return type depends on the chosen filter:
                - 'EQUALS_TRACK': Returns a list of TrackSegments.
                - 'EQUALS_OP': Returns an OperationalPoint.
                - 'EQUALS_TRACK_SEG': Returns a TrackSegment.
                - 'TRACKS_IN_BBOX': Returns a list of TrackSegments.
                - 'OP_IN_BBOX': Returns a list of OperationalPoints.
                - 'TRANSITIONS_IN_BBOX': Returns a list of TrackSegments.
                - 'BBOX': Returns a list of TrackSegments and OperationalPoints.
                - 'BOUNDARY': Returns a list of TrackSegments and OperationalPoints.

        Notes
        -----
        - The method validates the input parameters before creating the filter string.
        - Special characters in arguments are replaced by wildcards.
        - The method sends a request to the ISR API and returns the corresponding objects if the request is successful.
        - If the request fails or encounters an exception, an error is raised.
        """

        # check input
        self._check_input_for_query(filter_name=filter_name, args=args)

        # TRACKS
        if filter_name == 'EQUALS_TRACK':
            track_nr = args[0]
            names_and_track = []

            # query isr for json dict of all track segments
            response = self.query_isr(filter_name='EQUALS_TRACK', args=args)

            if response:
                track_segments = []
                # iterate over features
                for track_seg in response['features']:
                    # extract name and skip if name was already read (skips double entries)
                    # check for exceptional names
                    name = track_seg['properties']['ISR_STRECKE_VON_BIS']
                    if any([ex_name in name for ex_name in ISR_EXCEPTIONAL_STATION_NAMES]):
                        from_name, to_name = extract_station_names(track_name_string=name)
                        name = f'{from_name} - {to_name}'
                    name_ = name.replace('/', ' ')

                    km_type = 'raw' if enhance_kilometrage == False else 'enh'
                    file_path_1 = self.project_path / f'cache/track_segments/{track_nr}_{name}_{km_type}.pickle'
                    file_path_2 = (
                        self.project_path / f'cache/track_segments_continued/{track_nr}_{name}_{km_type}.pickle'
                    )

                    if file_path_1.exists():
                        track_segment = self._track_segment_from_pickle(file_path=file_path_1)

                    elif file_path_2.exists():
                        track_segment = self._track_segment_from_pickle(file_path=file_path_2)
                    else:
                        # query isr
                        name_ = name.replace('/', '*')

                        response = self.query_isr(filter_name='EQUALS_TRACK_SEG', args=[track_nr, name])
                        if response and len(response['features']) > 0:
                            track_segment = self._track_segment_from_feature_isr(
                                isr_feature=response['features'][0], enhance_kilometrage=enhance_kilometrage
                            )

                        else:
                            # request failed
                            # raise RequestException(f"An error occured while handling the request to ISR.")
                            continue

                    # skip double track segments (direction / opposite rail contains same geometry)
                    if (name, track_nr) in names_and_track and track_segment.properties[
                        'INF_GLEISANZAHL'
                    ] != 'eingleisig':
                        continue
                    names_and_track.append((name, track_nr))

                    track_segments.append(track_segment)

                return track_segments

            else:
                # request failed
                raise RequestException(f"An error occured while handling the request to ISR.")

        # TRACK SEGMENTS
        elif filter_name == 'EQUALS_TRACK_SEG':
            # filepath
            track_nr = args[0]
            name: str = args[1]  # type: ignore
            if any([ex_name in name for ex_name in ISR_EXCEPTIONAL_STATION_NAMES]):
                from_name, to_name = extract_station_names(track_name_string=name)
                name = f'{from_name} - {to_name}'
            name_ = name.replace('/', ' ')
            km_type = 'raw' if enhance_kilometrage == False else 'enh'
            file_path_1 = self.project_path / f'cache/track_segments/{track_nr}_{name}_{km_type}.pickle'
            file_path_2 = self.project_path / f'cache/track_segments_continued/{track_nr}_{name}_{km_type}.pickle'

            if file_path_1.exists():
                track_segment = self._track_segment_from_pickle(file_path=file_path_1)

            elif file_path_2.exists():
                track_segment = self._track_segment_from_pickle(file_path=file_path_2)

            else:
                # query isr
                args[1] = name.replace('/', '*')
                response = self.query_isr(filter_name='EQUALS_TRACK_SEG', args=args)

                if response:
                    track_segment = self._track_segment_from_feature_isr(
                        isr_feature=response['features'][0],
                        enhance_kilometrage=enhance_kilometrage,
                        cache_preprocessed_segments=cache_preprocessed_segments,
                    )
                else:
                    # request failed
                    raise RequestException(f"An error occured while handling the request to ISR.")

            return track_segment

        # OPERATIONAL POINTS
        elif filter_name == 'EQUALS_OP':
            track_nr = args[1]
            filepath = self.project_path / f'cache/operational_points/ops_track_{track_nr}.json'

            if not filepath.exists():
                # query ops for track and export json
                # have to query for betriebsstellen and streckenübergänge seperately
                # beacuse the API of ISR has changed and sending two queries in one
                # request is not working anymor
                response1 = self.query_isr(filter_name='EQUALS_OP', args=['*', track_nr])
                response2 = self.query_isr(filter_name='EQUALS_TRANSITION', args=['*', track_nr])

                if response1 and response2:
                    response = response1
                    response['features'] += response2['features']
                    self.export_json(response=response, filter_name='EQUALS_OP', args=['*', track_nr])
                else:
                    # request failed
                    raise RequestException(f"An error occured while handling the request to ISR.")

            operational_point = self._op_from_file(filepath, name=str(args[0]), track_nr=int(track_nr))

            return operational_point

        elif 'BBOX' in filter_name or 'BOUNDARY' in filter_name:
            # query isr for json dict of all track segments

            # have to query each feature type seperately (streckenabschnitte, betriebsstellen, streckenübergänge, tunnel)
            match filter_name:
                case 'BBOX':
                    subfilters = [filt for filt in self.filters if 'BBOX' in filt and filt != 'BBOX']
                case 'BOUNDARY':
                    subfilters = [filt for filt in self.filters if 'BOUNDARY' in filt and filt != 'BOUNDARY']
                case _:
                    subfilters = [filter_name]

            responses = []
            for filt in subfilters:
                response_ = self.query_isr(filter_name=filt, args=args)  # type: ignore
                responses.append(response_)

            if all(responses):
                logger.info('Query successful.')

                if len(responses) > 1:  # merge responses
                    response = responses[0]
                    for response_ in responses[1:]:
                        response['features'] += response_['features']

                names_and_track = []
                track_segments: list[TrackSegment] = []
                operational_points: list[OperationalPoint] = []
                transitions: list[OperationalPoint] = []
                # tunnels: list[Tunnel] TODO
                for feature in response['features']:
                    ID = feature['id']

                    if 'ISR_V_GEO_TEN_KLASSIFIZIERUNG' in ID:
                        # track segment

                        # extract attributes and format name (wildcard *)
                        track_nr = feature['properties']['ISR_STRE_NR']
                        name = feature['properties']['ISR_STRECKE_VON_BIS']
                        if any([ex_name in name for ex_name in ISR_EXCEPTIONAL_STATION_NAMES]):
                            from_name, to_name = extract_station_names(track_name_string=name)
                            name = f'{from_name} - {to_name}'
                        name = name.replace('/', '*')

                        km_type = 'raw' if enhance_kilometrage == False else 'enh'
                        filepath = self.project_path / f'cache/track_segments/{track_nr}_{name}_{km_type}.pickle'
                        matching_files = sorted(filepath.parent.glob(f'{track_nr}_{name}_{km_type}.pickle'))

                        if len(matching_files) == 1:
                            filepath = matching_files[0]
                            # load from file
                            track_segment = self._track_segment_from_pickle(file_path=filepath)
                            logger.info(f'Loaded from file: {track_nr} {name}.')

                        elif len(matching_files) > 1:
                            raise NotImplementedError('Not Implemented')

                        else:
                            # query isr
                            response = self.query_isr(filter_name='EQUALS_TRACK_SEG', args=[int(track_nr), str(name)])

                            if response:
                                logger.info(f'Preprocessing: {track_nr} {name}.')
                                track_segment = self._track_segment_from_feature_isr(
                                    isr_feature=response['features'][0],
                                    enhance_kilometrage=enhance_kilometrage,
                                    cache_preprocessed_segments=cache_preprocessed_segments,
                                )
                            else:
                                # request failed
                                raise RequestException(f"An error occured while handling the request to ISR.")

                        # skip double track segments (direction / opposite rail contains same geometry)
                        if (name, track_nr) in names_and_track and track_segment.properties[
                            'INF_GLEISANZAHL'
                        ] != 'eingleisig':
                            continue
                        names_and_track.append((name, track_nr))

                        track_segments.append(track_segment)

                    elif 'ISR_V_GEO_BETRIEBSSTELLEN_PUNKT' in ID:
                        # operational point
                        track_nr = feature['properties']['STRNR']
                        name = feature['properties']['BST_STELLE_NAME']
                        filepath = self.project_path / f'cache/operational_points/ops_track_{track_nr}.json'

                        if not filepath.exists():
                            # query ops for track and export json
                            response = self.query_isr(filter_name='EQUALS_OP', args=['*', track_nr])
                            if response:
                                self.export_json(response=response, filter_name='EQUALS_OP', args=['*', track_nr])
                            else:
                                # request failed
                                raise RequestException(f"An error occured while handling the request to ISR.")

                        operational_point = self._op_from_file(filepath, name=name, track_nr=int(track_nr))

                        operational_points.append(operational_point)

                    elif 'ISR_V_GEO_STRECKENUEBERGAENGE' in ID:
                        # operational point
                        track_nrs = [feature['properties']['STRECKE1'], feature['properties']['STRECKE2']]
                        name = feature['properties']['BST_STELLE_NAME']

                        for track_nr in track_nrs:
                            filepath = self.project_path / f'cache/operational_points/ops_track_{track_nr}.json'

                            if not filepath.exists():
                                # query ops for track and export json
                                response = self.query_isr(filter_name='EQUALS_OP', args=['*', track_nr])
                                if response:
                                    self.export_json(response=response, filter_name='EQUALS_OP', args=['*', track_nr])
                                else:
                                    # request failed
                                    raise RequestException(f"An error occured while handling the request to ISR.")

                            try:
                                transition = self._op_from_file(filepath, name=name, track_nr=int(track_nr))
                                transitions.append(transition)
                            except:
                                pass

                if 'TRACKS' in filter_name:
                    return track_segments

                elif 'OP' in filter_name:
                    return operational_points + transitions

                elif 'TRANSITION' in filter_name:
                    return transitions

                # elif 'TUNNEL' in filter_name:
                #    return tunnels

                else:  # all
                    return track_segments + operational_points + transitions  # + tunnels

            else:
                # request failed
                raise RequestException(f"An error occured while handling the request to ISR.")

        else:
            raise ValueError(f'Unknown filter: {filter_name}')

    def export_json(
        self,
        response: dict | None,
        filter_name: Literal[
            "EQUALS_TRACK",
            "EQUALS_OP",
            "EQUALS_TRACK_SEG",
            "TRACKS_IN_BBOX",
            "OP_IN_BBOX",
            "TRANSITIONS_IN_BBOX",
            "BBOX",
            "BOUNDARY",
        ],
        args: list[str | int | float],
        export_path: Path | None = None,
    ) -> dict | None:
        """
        Exports the response from an ISR query to a JSON file.

        Parameters
        ----------
        response: dict | None
            The response from the ISR query in JSON format. If None, raises a RequestException.

        filter_name: Literal
            The type of filter used in the ISR query. Choose from:
                - 'EQUALS_TRACK'
                - 'EQUALS_OP'
                - 'EQUALS_TRACK_SEG'
                - 'TRACKS_IN_BBOX'
                - 'OP_IN_BBOX'
                - 'TRANSITIONS_IN_BBOX'
                - 'BBOX'
                - 'BOUNDARY'

        args: list
            A list of arguments corresponding to the chosen filter. The interpretation of arguments depends on the selected filter.

        export_path: Path | None, optional
            The path to save the exported JSON file. If None, the default path is determined based on the filter and arguments.

        Returns
        -------
        None
            Exports the response to a JSON file at the specified path.

        Raises
        ------
        RequestException
            If the given response is None, indicating a failed request.

        ValueError
            If an unknown filter_name is provided.

        Notes
        -----
        - If export_path is not provided, it is determined based on the filter and arguments.
        - The exported JSON file is saved with utf-8 encoding and indented formatting for better readability.
        """

        if response:
            # set export path if no export path is given
            if not export_path:
                if filter_name == "EQUALS_TRACK":
                    export_path = self.project_path / f'cache/tracks/{args[0]}.json'

                elif filter_name == "EQUALS_TRACK_SEG":
                    export_path = self.project_path / f'cache/temp/track_segment_{args[0]}_{args[1]}.json'

                elif filter_name == 'EQUALS_OP':
                    if args[0] == '*':
                        export_path = self.project_path / f'cache/operational_points/ops_track_{args[1]}.json'
                    else:
                        export_path = self.project_path / f'cache/temp/{args[0]}_{args[1]}.json'

                elif filter_name == "TRACKS_IN_BBOX":
                    export_path = (
                        self.project_path / f'cache/temp/bbox_tracks_{args[0]}_{args[1]}_{args[2]}_{args[3]}.json'
                    )

                elif filter_name == "OP_IN_BBOX":
                    export_path = (
                        self.project_path / f'cache/temp/bbox_ops_{args[0]}_{args[1]}_{args[2]}_{args[3]}.json'
                    )

                elif filter_name == "TRANSITIONS_IN_BBOX":
                    export_path = (
                        self.project_path / f'cache/temp/bbox_transitions_{args[0]}_{args[1]}_{args[2]}_{args[3]}.json'
                    )

                elif filter_name == "BBOX":
                    export_path = self.project_path / f'cache/temp/bbox_{args[0]}_{args[1]}_{args[2]}_{args[3]}.json'

                elif filter_name == "BOUNDARY":
                    export_path = self.project_path / f'cache/temp/boundary_{args[0]}.json'

                else:
                    raise ValueError(f"Unkown Filter: {filter_name}")

            # save JSON
            with open(export_path, "w", encoding="utf-8") as f:
                json_dump(response, f, ensure_ascii=False, indent=4, sort_keys=True)

        else:
            # request failed
            raise RequestException(f"Given response is None. Export aborted.")

    def _check_input_for_query(
        self,
        filter_name: Literal[
            "EQUALS_TRACK",
            "EQUALS_OP",
            "EQUALS_TRACK_SEG",
            "TRACKS_IN_BBOX",
            "OP_IN_BBOX",
            "TRANSITIONS_IN_BBOX",
            "BBOX",
            "BOUNDARY",
        ],
        args: list[str | int | float],
    ):
        """
        Checks the validity of input parameters for an ISR query.

        Parameters
        ----------
        filter_name: Literal
            The type of filter to be applied. Choose from:
                - 'EQUALS_TRACK': Query by track. 'args' is a tuple of (<track_nr>).
                - 'EQUALS_OP': Query by operational point. 'args' is a tuple of (<name>, <track_nr>).
                - 'EQUALS_TRACK_SEG': Query by track segment. 'args' is a tuple of (<track_nr>, <name>).
                - 'TRACKS_IN_BBOX': Query tracks within a bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'OP_IN_BBOX': Query operational points within a bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'TRANSITIONS_IN_BBOX': Query transitions within a bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'BBOX': Query using a general bounding box. 'args' is a tuple of (<lon_min>, <lat_min>, <lon_max>, <lat_max>).
                - 'BOUNDARY': Query using a boundary (LinearRing). 'args' is a tuple of (<string of coordinates separated by space: "lon_1 lat_1 lon_2 lat_2 ... ">).

        args: list
            A list of arguments corresponding to the chosen filter. The interpretation of arguments depends on the selected filter.

        Raises
        ------
        ValueError
            If the filter_name is unknown.
            If the type of args is not a list.
            If the elements in args do not meet the expected types for the specified filter.
            If the number of elements in args is incorrect for the specified filter.
            If a 'BOUNDARY' filter is used, and args does not contain exactly one element: a string of coordinates separated by spaces.

        Notes
        -----
        - The method performs specific checks based on the filter_name to ensure the validity of input parameters for an ISR query.
        - Specific checks include verifying the types and number of elements in args based on the selected filter.
        - Raises a ValueError if any of the checks fail.
        """

        # check filter
        if not filter_name in self.filters:
            raise ValueError(f"Unknown filter: {filter_name}")
        # check type of args
        if not isinstance(args, list):
            raise ValueError(f"args must be of type list, passed type was {type(args)}")

        # check args for filter "EQUALS_TRACK"
        if filter_name == "EQUALS_TRACK":
            if not all([isinstance(val, (INT_TYPES, str)) for val in args]):
                raise ValueError(
                    f"When using filter 'EQUALS_TRACK', all elements in list args must be of typ int or str."
                )

        # check args for filter "EQUALS_OP"
        elif filter_name == "EQUALS_OP":
            if not isinstance(args[0], (INT_TYPES, str)):
                raise ValueError(
                    f"When using filter 'EQUALS_OP', first element in args must be of type str (name of operational point)."
                )
            if not isinstance(args[1], (INT_TYPES, str)):
                raise ValueError(
                    f"When using filter 'EQUALS_OP', second element in args must be of type int or str (track number)."
                )

        elif filter_name == "EQUALS_TRANSITION":
            if not isinstance(args[0], (INT_TYPES, str)):
                raise ValueError(
                    f"When using filter 'EQUALS_TRANSITION', first element in args must be of type str (name of transition)."
                )
            if not isinstance(args[1], (INT_TYPES, str)):
                raise ValueError(
                    f"When using filter 'EQUALS_TRANSITION', second element in args must be of type int or str (track number)."
                )

        # check args for filter "EQUALS_TRACK_SEG"
        elif filter_name == "EQUALS_TRACK_SEG":
            if not isinstance(args[0], (INT_TYPES, int)):
                raise ValueError(
                    f"When using filter 'EQUALS_TRACK_SEG', first element in args must be of type int (track number)."
                )
            if not isinstance(args[1], (INT_TYPES, str)):
                raise ValueError(
                    f"When using filter 'EQUALS_TRACK_SEG', second element in args must be of type str (track segment name)."
                )

        # check args for filters using "BBOX"
        elif "BBOX" in filter_name:
            if not len(args) == 4:
                raise ValueError(
                    "When using a 'BBOX' filter, the list args must contain exactly 4 elements: latitude_min, longitude_min, latitude_max and longitude_max."
                )

            if not all([isinstance(val, (FLOAT_TYPES, INT_TYPES)) for val in args]):
                raise ValueError(f"When using a 'BBOX' filter, all elements in list args must be of type float.")

        elif "BOUNDARY" in filter_name:
            if not (len(args) == 1) or type(args[0]) != str:
                raise ValueError(
                    "When using a 'BOUNDARY' filter, args must contain exactly one element: a string of coordinates seperated by spaces."
                )

            for val in args[0].split():
                try:
                    float(val)
                except:
                    raise ValueError(
                        "String coordinates can not be cast to float. Perhaps you used a comma instead of a point for decimals?"
                    )

        else:
            raise ValueError(f'Unknown filter: {filter_name}')

    # Operational Point methods

    @classmethod
    def _op_from_feature_isr(cls, isr_feature: dict) -> OperationalPoint:
        """
        Creates an OperationalPoint instance from ISR feature data.

        Parameters
        ----------
        isr_feature: dict
            The feature data obtained from the ISR API response.

        Returns
        -------
        OperationalPoint
            An instance of the OperationalPoint class representing the ISR feature.

        Raises
        ------
        ValueError
            If the feature type in the ISR data is unknown or unsupported.

        Notes
        -----
        - This method is used internally to create an OperationalPoint instance from ISR feature data.
        - It checks the feature type in the ISR data and extracts the relevant information.
        - The feature data is expected to contain 'id', 'properties', and 'geometry' keys.
        - The 'properties' key should contain a dictionary with property names as keys and their corresponding values as values.
        - The 'geometry' key should contain a dictionary with 'coordinates' key and the coordinates of the point.
        - The coordinates are transformed from WGS84 (EPSG:4326) to UTM (EPSG:31467).
        - The extracted information is used to create an OperationalPoint instance.
        """

        # extract feature id and properties
        feature_id = isr_feature['id']
        properties = dict(sorted(isr_feature['properties'].items()))

        # stations
        if "ISR_V_GEO_BETRIEBSSTELLEN_PUNKT" in feature_id:
            PROPERTIES = ISR_PROPERTIES_OPERATIONAL_POINT
            if isinstance(isr_feature['geometry'], type(None)):
                # project geo location from WGS84 (EPSG:4326) to UTM (EPSG:31467)
                point_wgs84 = Point(PROPERTIES['ALG_GEO_LAGE'][2](isr_feature['properties']['ALG_GEO_LAGE']))
                transformer = Transformer(source_cs='epsg:4326', target_cs='epsg:31467')
                point_epsg31467 = transformer.transform('source_to_target', Point((point_wgs84.y, point_wgs84.x)))
                # TODO: inverted (change in Transformer)
                coords_epsg31467 = point_epsg31467.y, point_epsg31467.x
            else:
                coords_epsg31467 = isr_feature['geometry']['coordinates']

        # track transitions
        elif "ISR_V_GEO_STRECKENUEBERGAENGE" in feature_id:
            PROPERTIES = ISR_PROPERTIES_JUNCTION
            coords_epsg31467 = isr_feature['geometry']['coordinates'][0]

        else:
            raise ValueError(f"Unknown element type: {feature_id}.")

        # create dict of type {'property': value}, where value is transformed from string to desired type
        properties = {
            key: l[2](value)
            for key, value, l in zip(PROPERTIES.keys(), properties.values(), PROPERTIES.values())
            if len(l) > 0
        }

        # create dict of type {'property': [property_description_short, property_description_long]}
        properties_info = {key: [l[0], l[1]] for key, l in zip(PROPERTIES.keys(), PROPERTIES.values()) if len(l) > 0}

        return OperationalPoint(coords_epsg31467, properties, properties_info)

    def _op_from_name_and_track_isr(
        self, name: str, track_nr: int, closest_to_geometry: Union[T, None] = None
    ) -> OperationalPoint:
        """
        Retrieves or creates an OperationalPoint instance based on the given name and track number.

        Parameters
        ----------
        name: str
            The name of the operational point.
        track_nr: int
            The track number associated with the operational point.
        closest_to_geometry: Union[T, None], optional
            An optional geometry (e.g., Point) used to find the closest OperationalPoint.
            If provided, the method calculates distances and selects the closest operational point.
            Default is None.

        Returns
        -------
        OperationalPoint
            An instance of the OperationalPoint class representing the specified operational point.

        Raises
        ------
        ValueError
            If no operational point is found with the given name and track number.

        Notes
        -----
        - This method queries the ISR API for operational points with the specified name and track number.
        - If there is a match, it creates an OperationalPoint instance from the ISR feature data.
        - If no match is found, it creates a placeholder OperationalPoint instance with limited information.
        - If there are multiple matches, the method selects the closest one based on the provided geometry or uses the first one if no geometry is provided.
        """

        # query ISR for operational points with matching name and track number

        if 'Leipzig-Semmelweiss' in name and track_nr == 6376:
            # station affiliated with 6376 and 6377 but in isr only exists with track nr 6377
            response = self.query_isr(filter_name="EQUALS_OP", args=[f"{name}", track_nr + 1])
            assert isinstance(response, dict)
            response['features'][0]['properties']['STRNR'] = 6376
        elif 'Köthen Süd' in name:
            response = self.query_isr(filter_name="EQUALS_OP", args=[f"{name}", 6403])
            assert isinstance(response, dict)
            response['features'][0]['properties']['STRNR'] = 6421
        else:
            response = self.query_isr(filter_name="EQUALS_OP", args=[f"{name}", track_nr])
            assert isinstance(response, dict)

        if len(response['features']) == 0:
            # empty properties dict for op's that don't exist in ISR
            isr_properties_op = ISR_PROPERTIES_OPERATIONAL_POINT
            keys = isr_properties_op.keys()
            properties = {key: [] for key in keys if len(isr_properties_op[key]) > 0}
            properties['STRNR'] = track_nr
            properties_info = {key: isr_properties_op[key] for key in keys if len(isr_properties_op[key]) > 0}

            if 'Frankfurt-Zehn Ruten' in name and len(name) == len('Frankfurt-Zehn Ruten'):
                properties['BST_STELLE_NAME'] = 'Frankfurt-Zehn Ruten'
                operational_point = OperationalPoint(
                    coords_epsg31467=(3474918, 5557118), properties=properties, properties_info=properties_info
                )

            elif 'Berlin-Spandau' in name and track_nr == 6107 and len(name) == len('Berlin-Spandau'):
                properties['BST_STELLE_NAME'] = 'Berlin-Spandau'
                operational_point = OperationalPoint(
                    coords_epsg31467=(3787180, 5830196), properties=properties, properties_info=properties_info
                )

            elif 'Wustermark Rbf Wot' in name and len(name) == len('Wustermark Rbf Wot'):
                properties['BST_STELLE_NAME'] = 'Wustermark Rbf Wot'
                operational_point = OperationalPoint(
                    coords_epsg31467=(3773494, 5831247), properties=properties, properties_info=properties_info
                )

            else:
                # no match: error
                raise ValueError(f"No operational point found with name = {name} for track = {track_nr}")

        elif len(response['features']) == 1:
            # one match: perfect

            operational_point = self._op_from_feature_isr(isr_feature=response['features'][0])

        else:
            # multiple match:
            # if 'closest_to_geometry' is given, compute nearest OperationalPoint to that geometry
            if closest_to_geometry:
                distances_list_m = [
                    closest_to_geometry.distance(Point(feature['geometry']['coordinates']))
                    if not isinstance(feature['geometry'], type(None))
                    else np.inf
                    for feature in response['features']
                ]
                operational_point = self._op_from_feature_isr(
                    isr_feature=response['features'][np.argmin(distances_list_m)]
                )
            # else take first OperationalPoint
            else:
                operational_point = self._op_from_feature_isr(isr_feature=response['features'][0])

        return operational_point

    def _op_from_file(
        self, filepath: Path, name: str, track_nr: int, closest_to_geometry: Union[T, None] = None
    ) -> OperationalPoint:
        """
        Retrieves or creates an OperationalPoint instance based on the given file path, name, and track number.

        Parameters
        ----------
        filepath: Path
            The file path to the JSON file containing the ISR features.
        name: str
            The name of the operational point.
        track_nr: int
            The track number associated with the operational point.
        closest_to_geometry: Union[T, None], optional
            An optional geometry (e.g., Point) used to find the closest OperationalPoint.
            If provided, the method calculates distances and selects the closest operational point.
            Default is None.

        Returns
        -------
        OperationalPoint
            An instance of the OperationalPoint class representing the specified operational point.

        Raises
        ------
        ValueError
            If no operational point is found with the given name and track number.

        Notes
        -----
        - This method reads the JSON file specified by the file path and searches for operational points with the given name and track number.
        - If there is a match, it creates an OperationalPoint instance from the ISR feature data.
        - If no match is found, it creates a placeholder OperationalPoint instance with limited information.
        - If there are multiple matches, the method selects the closest one based on the provided geometry or uses the first one if no geometry is provided.
        """

        # filepath must exist
        assert filepath.exists(), f'No file found for path: {filepath}'

        if 'Leipzig-Semmelweiss' in name and track_nr == 6376:
            # query ISR for operational points / junction with matching name and track number
            # TODO: check for file with track 6377
            # station affiliated with 6376 and 6377 but in isr only exists with track nr 6377
            response = self.query_isr(filter_name="EQUALS_OP", args=[f"{name}", track_nr + 1])
        else:
            response = {'features': []}
            # read file
            with open(filepath, encoding='utf-8') as f:
                json_dict = json.load(fp=f)
            # entries with fitting track and number
            for i in range(len(json_dict['features'])):
                json_feature = json_dict['features'][i]
                properties = json_feature['properties']

                if properties['BST_STELLENART'] == 'Stub':
                    equals_track = (
                        True
                        if int(properties['STRECKE1']) == track_nr or int(properties['STRECKE2']) == track_nr
                        else False
                    )
                else:
                    equals_track = True if int(properties['STRNR']) == track_nr else False

                if name == properties['BST_STELLE_NAME'] and equals_track == True:
                    response['features'].append(json_feature)
                elif 'Berlin-Moabit' in properties['BST_STELLE_NAME'] and track_nr == 6020:
                    response['features'].append(json_feature)

        assert isinstance(response, dict)

        if len(response['features']) == 0:
            # empty properties dict for op's that don't exist in ISR
            isr_properties_op = ISR_PROPERTIES_OPERATIONAL_POINT
            keys = isr_properties_op.keys()
            properties = {key: [] for key in keys if len(isr_properties_op[key]) > 0}
            properties_info = {key: isr_properties_op[key] for key in keys if len(isr_properties_op[key]) > 0}

            # operational points that are not included in ISR (only by name in track segment name)
            if 'Frankfurt-Zehn Ruten' in name and len(name) == len('Frankfurt-Zehn Ruten'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3474918, 5557118), properties=properties, properties_info=properties_info
                )

            elif 'Berlin-Spandau' in name and len(name) == len('Berlin-Spandau'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3787180, 5830196), properties=properties, properties_info=properties_info
                )

            elif 'Wustermark Rbf Wot' in name and len(name) == len('Wustermark Rbf Wot'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3773494, 5831247), properties=properties, properties_info=properties_info
                )

            elif 'Ohrstedt' in name and len(name) == len('Ohrstedt'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3514734, 6041428), properties=properties, properties_info=properties_info
                )

            elif 'Maschen Rbf, W 1219' in name and len(name) == len('Maschen Rbf, W 1219'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3568773, 5921582), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb1401_1500' in name and len(name) == len('StrUeb1401_1500'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3488287, 5883095), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb1401_1412' in name and len(name) == len('StrUeb1401_1412'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3485942, 5885359), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb1422_1425' in name and len(name) == len('StrUeb1422_1425'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3481499, 5890017), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb1570_1573' in name and len(name) == len('StrUeb1570_1573'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3385896, 5926982), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb1702_1703' in name and len(name) == len('StrUeb1702_1703'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3547178, 5807445), properties=properties, properties_info=properties_info
                )

            elif 'Bremerhaven-Speckenbüttel, W 5' in name and len(name) == len('Bremerhaven-Speckenbüttel, W 5'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3473047, 5940039), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb1630_1744' in name and len(name) == len('StrUeb1630_1744'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3487038, 5838159), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb2140_2146' in name and len(name) == len('StrUeb2140_2146'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3384633, 5706248), properties=properties, properties_info=properties_info
                )

            elif 'Essen-Steele, W 79' in name and len(name) == len('Essen-Steele, W 79'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3367746, 5702063), properties=properties, properties_info=properties_info
                )

            elif 'Essen-Steele, W 106' in name and len(name) == len('Essen-Steele, W 106'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3366687, 5702678), properties=properties, properties_info=properties_info
                )

            elif 'Gelsenkirchen Streckenw. 2172/2230' in name and len(name) == len(
                'Gelsenkirchen Streckenw. 2172/2230'
            ):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3365341, 5710866), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb2236_2246' in name and len(name) == len('StrUeb2236_2246'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3365980, 5713508), properties=properties, properties_info=properties_info
                )

            elif 'W 62' in name and len(name) == len('W 62'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3355085, 5671747), properties=properties, properties_info=properties_info
                )

            elif 'Wuppertal-Vohwinkel, W 88' in name and len(name) == len('Wuppertal-Vohwinkel, W 88'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3366224, 5679958), properties=properties, properties_info=properties_info
                )

            elif 'Vellmar-Obervellmar, W 16' in name and len(name) == len('Vellmar-Obervellmar, W 16'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3531788, 5691557), properties=properties, properties_info=properties_info
                )

            elif 'Rommerskirchen DB-Grenze' in name and len(name) == len('Rommerskirchen DB-Grenze'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3339604, 5658327), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb2650_2658' in name and len(name) == len('StrUeb2650_2658'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3360062, 5651676), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb2700_2705' in name and len(name) == len('StrUeb2700_2705'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3377964, 5673811), properties=properties, properties_info=properties_info
                )

            elif 'Meinerzhagen, W 3' in name and len(name) == len('Meinerzhagen, W 3'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3404502, 5664591), properties=properties, properties_info=properties_info
                )

            elif 'Bönen, W 51' in name and len(name) == len('Bönen, W 51'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3414759, 5720685), properties=properties, properties_info=properties_info
                )

            elif 'StrUeb1630_2982' in name and len(name) == len('StrUeb1630_2982'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3479662, 5821488), properties=properties, properties_info=properties_info  #
                )

            elif 'Ludwigshafen (Rh) BASF Südtor, W 529' in name and len(name) == len(
                'Ludwigshafen (Rh) BASF Südtor, W 529'
            ):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3459473, 5484127), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb3520_3687' in name and len(name) == len('StrUeb3520_3687'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3473778, 5548119), properties=properties, properties_info=properties_info  #
                )

            elif 'Schlüchtern, W 233' in name and len(name) == len('Schlüchtern, W 233'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3536336, 5578482), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb3683_3687' in name and len(name) == len('StrUeb3683_3687'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3474194, 5551324), properties=properties, properties_info=properties_info  #
                )

            elif 'Eppingen, W 61' in name and len(name) == len('Eppingen, W 61'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3495730, 5446222), properties=properties, properties_info=properties_info  #
                )

            elif 'Seckach, W 624' in name and len(name) == len('Seckach, W 624'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3524634, 5478312), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb4250_4661' in name and len(name) == len('StrUeb4250_4661'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3480155, 5310931), properties=properties, properties_info=properties_info  #
                )

            elif 'Abzw Lindau-Aeschach, W 2' in name and len(name) == len('Abzw Lindau-Aeschach, W 2'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3551608, 5269004), properties=properties, properties_info=properties_info  #
                )

            elif 'Stuttgart Hbf, W 117' in name and len(name) == len('Stuttgart Hbf, W 117'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3513838, 5405859), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb4701_4861' in name and len(name) == len('StrUeb4701_4861'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3513577, 5405533), properties=properties, properties_info=properties_info  #
                )

            elif 'Stuttgart-Zazenhausen, W 666 --Rbf--' in name and len(name) == len(
                'Stuttgart-Zazenhausen, W 666 --Rbf--'
            ):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3513299, 5412039), properties=properties, properties_info=properties_info  #
                )

            elif 'Stuttgart-Zazenhausen, W 655' in name and len(name) == len('Stuttgart-Zazenhausen, W 655'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3513299, 5412039), properties=properties, properties_info=properties_info  #
                )

            elif 'Stuttgart-Zazenhausen, W 655' in name and len(name) == len('Stuttgart-Zazenhausen, W 655'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3513299, 5412039), properties=properties, properties_info=properties_info  #
                )

            elif 'Stuttgart-Zazenhausen, W 656 --Rbf--' in name and len(name) == len(
                'Stuttgart-Zazenhausen, W 656 --Rbf--'
            ):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3513299, 5412039), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb4801_4861' in name and len(name) == len('StrUeb4801_4861'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3513585, 5405536), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb4812_4820' in name and len(name) == len('StrUeb4812_4820'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3513244, 5412066), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb5360_5361' in name and len(name) == len('StrUeb5360_5361'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3620449, 5324405), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb5420_5421' in name and len(name) == len('StrUeb5420_5421'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3552268, 5268420), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb5554_5604' in name and len(name) == len('StrUeb5554_5604'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3695500, 5336394), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb5556_5557' in name and len(name) == len('StrUeb5556_5557'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3703347, 5360591), properties=properties, properties_info=properties_info  #
                )

            elif 'Bft Regensburg Hafen Abzw, W 100' in name and len(name) == len('Bft Regensburg Hafen Abzw, W 100'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3729230, 5435556), properties=properties, properties_info=properties_info  #
                )

            elif 'Berlin Warschauer Straße, W 618' in name and len(name) == len('Berlin Warschauer Straße, W 618'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3802631, 5828460), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6007_6143' in name and len(name) == len('StrUeb6007_6143'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3806507, 5823149), properties=properties, properties_info=properties_info  #
                )

            elif 'Berlin-Schöneweide                S-Bahn, W 44' in name and len(name) == len(
                'Berlin-Schöneweide                S-Bahn, W 44'
            ):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3806507, 5823149), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6035_6172' in name and len(name) == len('StrUeb6035_6172'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3796418, 5821864), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6050_6346' in name and len(name) == len('StrUeb6050_6346'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3707554, 5710690), properties=properties, properties_info=properties_info  #
                )

            elif 'Halle Saalebrücke, W 8' in name and len(name) == len('Halle Saalebrücke, W 8'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3703809, 5705910), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6053_6346' in name and len(name) == len('StrUeb6053_6346'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3707536, 5708930), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6078_6525' in name and len(name) == len('StrUeb6078_6525'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3867136, 5839522), properties=properties, properties_info=properties_info  #
                )

            elif 'Strausberg                        S-Bahn, 75W53' in name and len(name) == len(
                'Strausberg                        S-Bahn, 75W53'
            ):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3828647, 5833109), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6081_6141' in name and len(name) == len('StrUeb6081_6141'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3798358, 5833578), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6109_6185' in name and len(name) == len('StrUeb6109_6185'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3786974, 5830222), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6110_6410' in name and len(name) == len('StrUeb6110_6410'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3686016, 5783410), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6118_6436' in name and len(name) == len('StrUeb6118_6436'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3692434, 5759143), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6126_6147' in name and len(name) == len('StrUeb6126_6147'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3810235, 5819898), properties=properties, properties_info=properties_info  #
                )

            elif 'Glasower Damm Süd, W 313' in name and len(name) == len('Glasower Damm Süd, W 313'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3800940, 5809833), properties=properties, properties_info=properties_info  #
                )

            elif 'Glasower Damm Süd, W 317' in name and len(name) == len('Glasower Damm Süd, W 317'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3800894, 5809989), properties=properties, properties_info=properties_info  #
                )

            elif 'Doberlug-Kirchhain Nord, 45W3' in name and len(name) == len('Doberlug-Kirchhain Nord, 45W3'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3816094, 5730576), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6142_6145' in name and len(name) == len('StrUeb6142_6145'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3809572, 5820264), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6145_6146' in name and len(name) == len('StrUeb6145_6146'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3809955, 5819092), properties=properties, properties_info=properties_info  #
                )

            elif 'Horka Gbf, W 9506' in name and len(name) == len('Horka Gbf, W 9506'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3809955, 5819092), properties=properties, properties_info=properties_info  #
                )

            elif 'Knappenrode, W 1' in name and len(name) == len('Knappenrode, W 1'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3871843, 5711437), properties=properties, properties_info=properties_info  #
                )

            elif 'Dresden-Neustadt, W 726' in name and len(name) == len('Dresden-Neustadt, W 726'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3832075, 5669419), properties=properties, properties_info=properties_info  #
                )

            elif 'Dresden Hbf, W 244' in name and len(name) == len('Dresden Hbf, W 244'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3832487, 5666543), properties=properties, properties_info=properties_info  #
                )

            elif 'Dresden Freiberger Straße, W 526' in name and len(name) == len('Dresden Freiberger Straße, W 526'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3830931, 5667798), properties=properties, properties_info=properties_info  #
                )

            elif 'Dresden Mitte, W 562' in name and len(name) == len('Dresden Mitte, W 562'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3831337, 5668691), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6243_6258' in name and len(name) == len('StrUeb6243_6258'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3830740, 5667075), properties=properties, properties_info=properties_info  #
                )

            elif 'Dresden-Altstadt, W 119' in name and len(name) == len('Dresden-Altstadt, W 119'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3830740, 5667075), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6248_6251' in name and len(name) == len('StrUeb6248_6251'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3821607, 5675481), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6248_6274' in name and len(name) == len('StrUeb6248_6274'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3830259, 5668401), properties=properties, properties_info=properties_info  #
                )

            elif 'Glauchau-Schönbörnchen, W 42' in name and len(name) == len('Glauchau-Schönbörnchen, W 42'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3746681, 5637534), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6274_6363' in name and len(name) == len('StrUeb6274_6363'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3810873, 5691604), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6321_6954' in name and len(name) == len('StrUeb6321_6954'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3793377, 6050754), properties=properties, properties_info=properties_info  #
                )

            elif 'Abzw Plaaz, 66W 1' in name and len(name) == len('Abzw Plaaz, 66W 1'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3719554, 5972423), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6325_6929' in name and len(name) == len('StrUeb6325_6929'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3706058, 5996997), properties=properties, properties_info=properties_info  #
                )

            elif 'Halle (Saale) Hbf, 92W224' in name and len(name) == len('Halle (Saale) Hbf, 92W224'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3707435, 5709289), properties=properties, properties_info=properties_info  #
                )

            elif 'Halle (Saale) Gbf, 95W446' in name and len(name) == len('Halle (Saale) Gbf, 95W446'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3707524, 5709397), properties=properties, properties_info=properties_info  #
                )

            elif 'Halle (Saale) Gbf Nord, 95W408' in name and len(name) == len('Halle (Saale) Gbf Nord, 95W408'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3709125, 5711250), properties=properties, properties_info=properties_info  #
                )

            elif 'Halle (Saale) Gbf, 96W460' in name and len(name) == len('Halle (Saale) Gbf, 96W460'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3708829, 5707515), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6376_6377' in name and len(name) == len('StrUeb6376_6377'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3736342, 5692099), properties=properties, properties_info=properties_info  #
                )

            elif 'Leipzig-Semmelweisstraße, 97W193' in name and len(name) == len('Leipzig-Semmelweisstraße, 97W193'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3736342, 5692099), properties=properties, properties_info=properties_info  #
                )

            elif 'Köthen Süd, 05W507' in name and len(name) == len('Köthen Süd, 05W507'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3707407, 5736537), properties=properties, properties_info=properties_info  #
                )

            elif 'Abzw Halle (Saale) Leuchtturm, W 6' in name and len(name) == len(
                'Abzw Halle (Saale) Leuchtturm, W 6'
            ):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3709376, 5707046), properties=properties, properties_info=properties_info  #
                )

            elif 'Halle (Saale) Hbf, W 355' in name and len(name) == len('Halle (Saale) Hbf, W 355'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3707606, 5708948), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6403_6434' in name and len(name) == len('StrUeb6403_6434'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3679821, 5779401), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6422_6436' in name and len(name) == len('StrUeb6422_6436'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3692866, 5760802), properties=properties, properties_info=properties_info  #
                )

            elif 'Rostock Gbf, Gl 8' in name and len(name) == len('Rostock Gbf, Gl 8'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3706024, 5997711), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6785_6942' in name and len(name) == len('StrUeb6785_6942'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3771146, 5921609), properties=properties, properties_info=properties_info  #
                )

            elif 'StrUeb6953_6954' in name and len(name) == len('StrUeb6953_6954'):
                operational_point = OperationalPoint(
                    coords_epsg31467=(3795593, 6050510), properties=properties, properties_info=properties_info  #
                )

            else:
                # no match: error
                raise ValueError(f"No operational point found with name = {name} for track = {track_nr}")

        elif len(response['features']) == 1:
            # one match: perfect

            operational_point = self._op_from_feature_isr(isr_feature=response['features'][0])

        else:
            # multiple match:
            # if 'closest_to_geometry' is given, compute nearest OperationalPoint to that geometry
            if closest_to_geometry:
                distances_list_m = [
                    closest_to_geometry.distance(Point(feature['geometry']['coordinates']))
                    if not isinstance(feature['geometry'], type(None))
                    else np.inf
                    for feature in response['features']
                ]
                operational_point = self._op_from_feature_isr(
                    isr_feature=response['features'][np.argmin(distances_list_m)]
                )
            # else take first OperationalPoint
            else:
                operational_point = self._op_from_feature_isr(isr_feature=response['features'][0])

        return operational_point

    # Track Segment methods

    def _track_segment_from_pickle(self, file_path: Path) -> TrackSegment:
        """
        Read a TrackSegment from a pickle file.

        Parameters
        ----------
        file_path : Path
            The file path to the pickle file containing the serialized TrackSegment.

        Returns
        -------
        TrackSegment
            An instance of the TrackSegment class read from the pickle file.

        Raises
        ------
        ValueError
            If the given filepath does not exist or does not point to a pickle file.

        Notes
        -----
        - This method reads a TrackSegment instance serialized in a pickle file.
        - The file path must exist, and the file extension must be '.pickle'.
        - If successful, it returns the deserialized TrackSegment instance.
        - If the file path is invalid or the file is not a pickle file, a ValueError is raised.
        """

        if file_path.exists() and file_path.suffix == '.pickle':
            with open(file_path, "rb") as file:
                track_segment: TrackSegment = pickle.load(file)

            return track_segment

        else:
            raise ValueError(f'Given filepath does not exist or does not point to a pickle file: {file_path}')

    def _track_segment_from_track_and_name(
        self, track: int, name: str, allow_previous: bool = True, cache_preprocessed_segments: bool = True
    ) -> TrackSegment:
        """
        Create a TrackSegment instance from a track number and name.

        Parameters
        ----------
        track : int
            The track number for which the TrackSegment is to be created.
        name : str
            The name of the track segment.
        allow_previous : bool, optional
            If True, allows an TrackSegment instance to load previous track segments (used to compute kilometrage offset). If False, it does not allow. Default value should be used, which is True.

        Returns
        -------
        TrackSegment
            An instance of the TrackSegment class created from the specified track number and name.

        Notes
        -----
        - This method checks the cache for a previously saved TrackSegment.
        - If found in the cache, it loads and returns the TrackSegment from the pickle file.
        - If not found, it queries the ISR API for the specified track and name to create a new TrackSegment.
        """

        file_path_1 = self.project_path / f'cache/track_segments/{track}_{name}.pickle'
        file_path_2 = self.project_path / f'cache/track_segments_continued/{track}_{name}.pickle'

        if file_path_1.exists():
            track_segment = self._track_segment_from_pickle(file_path=file_path_1)

        elif file_path_2.exists():
            track_segment = self._track_segment_from_pickle(file_path=file_path_2)

        else:
            response = self.query_isr(filter_name='EQUALS_TRACK_SEG', args=[track, name])
            assert isinstance(response, dict)
            track_segment = self._track_segment_from_feature_isr(
                isr_feature=response,
                allow_previous=allow_previous,
                cache_preprocessed_segments=cache_preprocessed_segments,
            )

        return track_segment

    def _track_segment_from_feature_isr(
        self,
        isr_feature: dict,
        allow_previous: bool = True,
        enhance_kilometrage: bool = True,
        cache_preprocessed_segments: bool = True,
    ) -> TrackSegment:
        """
        Create a TrackSegment instance from ISR response (JSON dictionary).

        Parameters
        ----------
        isr_feature : dict
            The ISR response containing information about the track segment in JSON dictionary format.
        allow_previous : bool, optional
            If True, allows an TrackSegment instance to load previous track segments (used to compute kilometrage offset). If False, it does not allow. Default value should be used, which is True.

        Returns
        -------
        TrackSegment
            An instance of the TrackSegment class created from the ISR response.

        Notes
        -----
        - This method extracts relevant information from the ISR response to create a TrackSegment instance.
        - It checks if the feature in the response corresponds to a track segment based on its feature ID.
        - It parses the properties and creates a dictionary of track segment properties with their corresponding values.
        - It retrieves operational points (stations) associated with the start and end of the track segment.
        - It preprocesses the track segment lines using an ISR preprocessor.
        - It creates a TrackSegment instance with processed lines, properties, and operational points.
        """

        # extract feature id and properties
        feature_id = isr_feature['id']
        properties = dict(sorted(isr_feature['properties'].items()))
        track_nr = int(properties['ISR_STRE_NR'])
        name = properties['ISR_STRECKE_VON_BIS']
        track_type = properties['INF_GLEISANZAHL']

        if any([ex_name in name for ex_name in ISR_EXCEPTIONAL_STATION_NAMES]):
            from_name, to_name = extract_station_names(track_name_string=name)
            name_ = f'{from_name} - {to_name}'
            name_ = name_.replace('/', '*')
        else:
            name_ = name

        # extract properties for opposite track (if double track)
        if track_type != 'eingleisig':
            # query with same track nr and name
            response = self.query_isr('EQUALS_TRACK_SEG', args=[track_nr, name_])
            n_elements = len(response['features'])  # number of found elements

            if n_elements == 0:
                raise ValueError('No elements found')
            elif n_elements == 1:
                raise ValueError(f'Found only one track but track type is {track_type}')
            else:
                for element in response['features']:
                    if (
                        element['properties']['ISR_STRECKE_VON_BIS'] == name
                        and element['properties']['INF_GLEISANZAHL'] != track_type
                    ):
                        properties_2 = dict(sorted(element['properties'].items()))
                        break

                else:
                    properties_2 = None

        else:
            properties_2 = None

        # set var 'properties' for Richtungsgleis, and var 'properties_2' for Gegengleis
        # for monorails, only use 'properties', as 'properties_2' is None
        if properties_2:
            if track_type == 'Gegengleis':
                temp = deepcopy(properties)
                properties = deepcopy(properties_2)
                properties_2 = deepcopy(temp)

        # assert feature is a track segment
        assert "ISR_V_GEO_TEN_KLASSIFIZIERUNG" in feature_id, f"Unknown element type: {feature_id}."

        # set properties
        PROPERTIES = ISR_PROPERTIES_TRACK_SEGMENTS

        # parse properties
        properties = {
            key: l[2](value)
            for key, value, l in zip(PROPERTIES.keys(), properties.values(), PROPERTIES.values())
            if len(l) > 0
        }

        if properties_2:
            properties_2 = {
                key: l[2](value)
                for key, value, l in zip(PROPERTIES.keys(), properties_2.values(), PROPERTIES.values())
                if len(l) > 0
            }

        # create dict of type {'property': [property_description_short, property_description_long]}
        properties_info = {key: (l[0], l[1]) for key, l in zip(PROPERTIES.keys(), PROPERTIES.values()) if len(l) > 0}

        # get coordinates
        isr_original_lines_epsg31467 = MultiLineString(isr_feature['geometry']['coordinates'])

        # name of file with ops for track
        track_ops_file_path = (
            Path(__file__).parent.parent.parent.parent / f'cache/operational_points/ops_track_{track_nr}.json'
        )

        if not track_ops_file_path.exists():
            # query all operational points for track and write to cache
            response1 = self.query_isr(filter_name='EQUALS_OP', args=['*', track_nr])
            response2 = self.query_isr(filter_name='EQUALS_TRANSITION', args=['*', track_nr])
            response = response1
            response['features'] += response2['features']
            assert isinstance(response, dict)
            self.export_json(response=response, filter_name='EQUALS_OP', args=['*', track_nr])

        # get names of stations at start and end of track segment
        from_name, to_name = extract_station_names(track_name_string=name, split_only=True)

        # read operational points from cache
        operational_point_from = self._op_from_file(
            filepath=track_ops_file_path,
            name=from_name,
            track_nr=track_nr,
            closest_to_geometry=isr_original_lines_epsg31467,
        )
        operational_point_to = self._op_from_file(
            filepath=track_ops_file_path,
            name=to_name,
            track_nr=track_nr,
            closest_to_geometry=isr_original_lines_epsg31467,
        )

        # preprocess track segment lines
        isr_preprocessor = ISRPreprocessor(
            operational_point_from=operational_point_from, operational_point_to=operational_point_to
        )
        lines_processed = isr_preprocessor.process_track_segment_lines(lines=list(isr_original_lines_epsg31467.geoms))

        # create instance
        track_segment = TrackSegment(
            lines_processed,
            properties,
            properties_2,
            properties_info,
            operational_point_from,
            operational_point_to,
            enhance_kilometrage,
            allow_previous,
            cache_preprocessed_segments,
        )

        return track_segment



import numpy as np
from isr_matcher.geometry.map import Map
from isr_matcher.geometry.rail import Rail
from isr_matcher.geometry.track_segment import TrackSegment
from shapely.geometry import Point, LineString
import numpy as np
from isr_matcher.ops._functions import angle_between, split_line_at_point
from isr_matcher._constants._parse import db_km_to_km
from isr_matcher.data_handlers.transformer import Transformer
from shapely.ops import nearest_points
from typing import Tuple, Literal
from itertools import groupby
from isr_matcher.geometry.gnss_series import GNSSSeries
import bisect
import folium
import pandas as pd
from pathlib import Path
from scipy.interpolate import splprep, splev
from itertools import cycle

from isr_matcher._constants.logging import setup_logger
import logging

# Create logger for the current module
setup_logger()
logger = logging.getLogger(__name__)


class MMPostprocessor:
    """Postprocesses the results of a map matching..

    Attributes
    ----------
    _map : Map
        The map.

    """

    def __init__(self, map_: Map):
        self._map = map_

    @property
    def map(self) -> Map:
        """
        Initializes an MMPostprocessor object with the specified map.

        Parameters
        ----------
        map_ : Map
            The map.
        """
        return self._map

    def concatenate_rails(
        self, S: list[int], start_point: Point, end_point: Point, resolution_m: float = 5
    ) -> LineString:
        """Concatenates rail segments to form a continuous path.

        Parameters
        ----------
        S : list[int]
            List of rail indices.
        start_point : Point
            Starting point of the path.
        end_point : Point
            Ending point of the path.
        resolution_m : float, optional
            Resolution for spline interpolation in meters, by default 5.

        Returns
        -------
        LineString
            A LineString representing the concatenated rail segments forming a continuous path.
        """

        path_by_index: list[int] = np.array(S)[np.concatenate(([True], np.diff(S) != 0))].tolist()
        path_by_rails: list[Rail] = [self.map['rails'][s] for s in path_by_index]

        # concatenate rails
        path = path_by_rails[0]

        for i in range(1, len(path_by_rails)):
            # special case
            if (
                path_by_rails[i].track_segment_name
                in 'Ludwigshafen (Rhein) Hbf, W 14 - Ludwigshafen (Rhein) Mitte 3401'
                and path_by_rails[i - 1].track_segment_name
                in 'Ludwigshafen (Rhein) Überleitung Süd - Ludwigshafen (Rhein) Mitte 3522'
            ):
                continue

            no_split = False
            split_point = nearest_points(path_by_rails[i], Point(path.coords[-1]))[0]
            split_postion = path_by_rails[i].project(split_point, normalized=True)
            split_postion_pre = path_by_rails[i - 1].project(split_point, normalized=True)

            # clip next rail (if transition at middle of rail, remove part of rail before transition point)
            if split_postion > 0 and split_postion < 1:
                pre_split_point = nearest_points(path_by_rails[i], Point(path.coords[-2]))[0]
                distance = path_by_rails[i].project(split_point) - path_by_rails[i].project(pre_split_point)

                normal_distance = split_point.distance(path_by_rails[i - 1])
                if distance > 0:
                    new_split_point = path_by_rails[i].interpolate(
                        path_by_rails[i].project(split_point) + normal_distance * 6
                    )
                elif distance < 0:
                    new_split_point = path_by_rails[i].interpolate(
                        path_by_rails[i].project(split_point) - normal_distance * 6
                    )
                else:
                    segment = path_by_rails[i]
                    no_split = True
                if no_split == False:
                    segment1, segment2 = split_line_at_point(path_by_rails[i], new_split_point)

                    if distance > 0:
                        segment = segment2
                    else:
                        segment = segment1

            # check if next segment has opposite direction
            # if it has, skip segment if next segment after is very close (to avoid running backwards)
            elif (split_postion <= 1e-3 and split_postion_pre >= 1 - 1e-3) or (
                split_postion_pre <= 1e-3 and split_postion >= 1 - 1e-3
            ):
                if path_by_rails[i - 1].distance(path_by_rails[i + 1]) <= 10:
                    continue
                else:
                    segment = path_by_rails[i]

            else:
                segment = path_by_rails[i]

            path = MMPostprocessor.join(path, segment)

        # clip path (if path starts/ends at middle of rail, remove part of rail after/before)
        start_point_on_path = nearest_points(path, start_point)[0]
        try:
            p1, p2 = split_line_at_point(path, start_point_on_path)
            if p1.distance(nearest_points(path, end_point)[0]) <= p2.distance(nearest_points(path, end_point)[0]):
                path = p1
            else:
                path = p2
        except ValueError:
            pass
        end_point_on_path = nearest_points(path, end_point)[0]
        try:
            p1, p2 = split_line_at_point(path, end_point_on_path)
            if p1.distance(start_point_on_path) <= p2.distance(start_point_on_path):
                path = p1
            else:
                path = p2
        except ValueError:
            pass

        ### smooth path (spline interpolation)
        try:
            # remove hard turns from path
            index = []
            for i in range(1, len(path.coords) - 1):
                v1 = np.array(path.coords[i]) - np.array(path.coords[i - 1])
                v2 = np.array(path.coords[i + 1]) - np.array(path.coords[i])
                a = angle_between(v1, v2)
                if abs(a) > 60:
                    index.append(i)
            x = np.array([c[0] for c in path.coords])
            y = np.array([c[1] for c in path.coords])
            x = np.delete(x, index)
            y = np.delete(y, index)
            path = LineString(np.column_stack([x, y]))

            # generate spline points by interpolating along path (1 point per 50 meter)
            # if spline points are too densely sampled from path, the spline will follow
            # the linear path, which makes spline interpolation redundant
            L = path.length
            interpol_points = path.interpolate(np.linspace(0, 1, int(L / 50)), normalized=True)
            x = np.array([c.x for c in interpol_points])
            y = np.array([c.y for c in interpol_points])

            # create spline
            points = np.column_stack((y, x))
            tck, u = splprep(points.T, s=0)

            # evaluate spline (1 point per 'resoltion_m' meter)
            y, x = splev(np.linspace(0, 1, int(L / resolution_m)), tck)
            path = LineString(np.column_stack([x, y]))

        except:
            pass

        return path

    @staticmethod
    def join(line: LineString | Rail, other: LineString | Rail) -> LineString:
        """Concatenates two LineString or Rail instances into one LineString.

        Parameters
        ----------
        line: LineString | Rail
            First LineString or Rail instance.
        other: LineString | Rail
            Second LineString or Rail instance.

        Returns
        -------
        LineString
            Concatenated LineString formed by joining the two input LineStrings or Rails.
        """

        line1 = line
        line2 = other

        # ensure line coordinates are ordered in same direction
        d1s2s, d1s2n = Point(line1.coords[0]).distance([Point(line2.coords[0]), Point(line2.coords[-1])])
        d1n2s, d1n2n = Point(line1.coords[-1]).distance([Point(line2.coords[0]), Point(line2.coords[-1])])
        d_min = np.min([d1s2s, d1s2n, d1n2s, d1n2n])

        if d_min == d1s2s:  # line1 reversed, line2 normal
            line1 = line1.reverse()
        elif d_min == d1s2n:  # line1 reversed, line2 reversed
            line1 = line1.reverse()
            line2 = line2.reverse()
        elif d_min == d1n2s:  # line1 normal, line2 normal
            pass
        else:  # line1 normal, line2 reversed
            line2 = line2.reverse()

        # if lines are connected already, skip connecting
        d = Point(line1.coords[-1]).distance(Point(line2.coords[0]))
        if d < 1e-6:
            return LineString(list(line1.coords) + list(line2.coords[1:]))

        else:
            return LineString(list(line1.coords) + list(line2.coords))

    def compute_position_parameter(self, path: LineString, rail_sequence: list[Rail], gnss_coords: list[Point]):
        """
        Fucntion to compute positional parameters in track space (track_nr, dircetion, kilometrage and projected gnss coordinates).

        Parameters
        ----------
        path: LineString
            LineString representing the continous path of vehicle.
        rail_sequence: list[Rail]
            List of rails. Each entry is the assigned rail for corresponding GNSS measurment.
        gnss_coords: list[Point]
            The GNSS measurements in UTM coordinate system (EPSG:31467).

        Note
        ----
        The lists 'rail_sequence' and 'gnss_coords' must be of equal length. Each rail corresponds to exactly one measurement, therefore Rails can be repeated. Same for 'track_segment_sequence'.

        """

        # compute kilometrage and projections on path
        matched_points_utm = []
        matched_points_wgs = []
        kms_db = []
        kms = []
        kms_running = [0.0]
        directions = []
        rails = []
        gnss_error_m = []
        inclines = []
        for i, (opt_rail, point) in enumerate(zip(rail_sequence, gnss_coords)):
            # projection on path
            matched_point_utm = nearest_points(path, point)[0]
            matched_points_utm.append(matched_point_utm)

            # gnss error
            error_m = point.distance(matched_point_utm)
            gnss_error_m.append(error_m)

            # projection of matched point (utm to wgs)
            transformer = Transformer(source_cs='epsg:31467', target_cs='epsg:4326')
            matched_point_wgs = transformer.transform(
                'source_to_target', Point([matched_point_utm.y, matched_point_utm.x])
            )
            matched_points_wgs.append(matched_point_wgs)

            # compute parameters
            for j, track_segment in enumerate(self.map['track_segments']):
                # find matching track segment for rail
                if (int(opt_rail.track) == int(track_segment.track_nr)) and (
                    opt_rail.track_segment_name == track_segment.name
                ):
                    # get km
                    if track_segment.kilometrage.n_lines == 2 and opt_rail.direction == 0:
                        if len(directions) > 0 and not np.isnan(directions[-1]):
                            direction = directions[-1]
                        else:
                            direction = 1
                    else:
                        direction = opt_rail.direction
                    km_db = track_segment.km(point=matched_point_utm, direction=direction, return_type='db')
                    km = db_km_to_km(km_db)
                    kms_db.append(km_db)
                    kms.append(km)

                    # get rail types
                    match opt_rail.direction:
                        case 0:
                            rail_type = 'eingleisig'
                        case 1:
                            rail_type = 'Richtungsgleis'
                        case 2:
                            rail_type = 'Gegengleis'
                        case _:
                            raise ValueError('Unknown rail direction.')
                    rails.append(rail_type)

                    # running km: distances are computed based on path geometry, not by kilometrage values
                    if i > 0:
                        # running km
                        km_running = (
                            abs(path.project(matched_points_utm[i]) - path.project(matched_points_utm[i - 1])) / 1000
                        )

                        kms_running.append(kms_running[-1] + km_running)

                        # the rest just to get the direction...
                        # same track segment as for last gnss point / last km
                        if rail_sequence[i] == rail_sequence[i - 1]:
                            # running distance
                            distance_since_last_point_km = (
                                opt_rail.project(gnss_coords[i]) - opt_rail.project(gnss_coords[i - 1])
                            ) / 1000  # abs(track_segment.km(point=gnss_coords[i], direction=int(direction), return_type='float') - track_segment.km(point=gnss_coords[i-1], direction=int(direction), return_type='float'))
                            # abs_distance_since_last_point_km = abs(distance_since_last_point_km)

                            # direction
                            if distance_since_last_point_km > 0:
                                direction = 1
                            elif distance_since_last_point_km < 0:
                                direction = 2
                            else:
                                direction = directions[-1]  # use last direction

                        # rail changed since last gnss point / last km
                        else:
                            # distance from last point to current rail end (rail end can also be start of rail)
                            distance_rail_start, distance_rail_end = gnss_coords[i - 1].distance(
                                [Point(opt_rail.coords[0]), Point(opt_rail.coords[-1])]
                            )
                            if distance_rail_start < distance_rail_end:
                                # ride on rail started at beginning of rail
                                rail_end = Point(opt_rail.coords[0])
                                distance_2 = (opt_rail.project(gnss_coords[i]) - opt_rail.project(rail_end)) / 1000

                            else:
                                # ride on rail started at end of rail
                                rail_end = Point(opt_rail.coords[-1])
                                distance_2 = -(opt_rail.project(rail_end) - opt_rail.project(gnss_coords[i])) / 1000

                            # direction: (we use moving direction on current rail and ignore direction on last rail)
                            if distance_2 > 0:
                                direction = 1
                            elif distance_2 < 0:
                                direction = 2
                            else:
                                direction = directions[-1]  # use last direction

                    else:
                        # set initial direction to nan because it can not be determined
                        # will be set to the next direction later
                        direction = np.nan

                    # get incline
                    incline = track_segment.incline(km=km, direction=opt_rail.direction)
                    inclines.append(incline)

                    # skip to next measurement
                    directions.append(direction)
                    break

            else:
                raise ValueError('No matching track segment found.')

        # replace nan values by next neighbor value
        directions = [
            directions[i] if np.isnan(directions[i]) == False else directions[i + 1]
            for i in range(len(directions))[::-1]
        ][::-1]

        position_parameter_dict = {
            'track_numbers': [int(rail.track) for rail in rail_sequence],
            'rail_types': rails,
            'directions': directions,
            'kms': kms,
            'kms_db': kms_db,
            'kms_running': kms_running,
            'matched_coordinates_utm': matched_points_utm,
            'matched_coordinates_wgs': matched_points_wgs,
            'gnss_error_m': gnss_error_m,
            'inclines': inclines,
        }

        return position_parameter_dict

    def compute_track_segment_sequence(self, rail_sequence: list[Rail]) -> Tuple[list[TrackSegment], list[int]]:
        """Computes the track segment sequence given a sequence of rails.

        Parameters
        ----------
        rail_sequence: list[Rail]
            A sequence of rails.

        Returns
        -------
        Tuple[list[TrackSegment], list[int]]
            A tuple containing the list of track segments corresponding to the input rail sequence
            and the indices of the rails in the input sequence that match the track segments.

        Raises
        ------
        ValueError
            If a rail cannot be matched with a track segment.
        """

        # lists
        track_segment_sequence = []
        track_segment_indices = []
        for i, rail in enumerate(rail_sequence):
            # find rail in track segments
            for track_segment in self.map['track_segments']:
                # set skip bool
                skip = False
                # skip if previous rail is the same as current rail
                if (i == 0) or (rail_sequence[i - 1].track_segment_name != rail_sequence[i].track_segment_name):
                    if (int(rail.track) == int(track_segment.track_nr)) and (
                        rail.track_segment_name == track_segment.name
                    ):
                        # append track segment to list
                        track_segment_sequence.append(track_segment)
                        track_segment_indices.append(i)
                        break

                else:
                    skip = True

            else:
                if skip == False:
                    raise ValueError('Could not match rail with track segment')

        return track_segment_sequence, track_segment_indices

    def compute_incline_profile(
        self,
        path: LineString,
        gnss_coords_utm: list[Point],
        track_segment_sequence: list[TrackSegment],
        track_segment_indices: list[int],
        rail_sequence: list[Rail],
        position_dict: dict,
    ):
        """Computes the incline profile along a given path based on railway infrastructure data and GNSS coordinates.

        Parameters
        ----------
        path : LineString
            The path along which to compute the incline profile.
        gnss_coords_utm : list[Point]
            The GNSS coordinates as a list of Point objects.
        track_segment_sequence : list[TrackSegment]
            The sequence of track segments corresponding to the rail sequence.
        track_segment_indices : list[int]
            The indices of the rails in the input sequence that match the track segments.
        rail_sequence : list[Rail]
            The sequence of rails.
        position_dict : dict
            A dictionary containing position information.

        Returns
        -------
        dict
            A dictionary containing the incline profile with keys 'kms' for the distances in kilometers,
            'inclines' for the corresponding inclines, and 'track_segment_names' for the names of the track segments.

        Raises
        ------
        ValueError
            If the direction cannot be determined.
        """

        # inclines at kilometrage values

        inclines_list = []
        kms_inclines_list = []
        track_segment_names_list = []
        track_segment_indices_new = track_segment_indices + [len(rail_sequence) - 1]
        for n, (i1, i2) in enumerate(zip(track_segment_indices_new[:-1], track_segment_indices_new[1:])):
            # i1:i2 -> rails belonging to a single track segment, i1: first occurence of a rail belonging track segment, i2-1: last occurence of a rail belonging track segment

            # it is possible that the train changes movement direction while driving on the track segment
            # if moving dircetion = 2 (in direction of decreasing kilometrage), inclines must be negated and reversed.
            # therefore, we have to check if train changes direction while on the track segment and if so,
            # split the ride into trips where each trips has a single direction
            moving_directions = position_dict['directions']
            trips = [list(g) for k, g in groupby(moving_directions[i1:i2])]

            # kms / inclines per trip
            # transition between trips is instaneous (last km_val+incline of last trip jumps to first km_val/incline of next trip)
            trips_kms = []
            trips_inclines = []
            track_segment_names = []
            for t in range(len(trips)):
                # trip indices
                ti = sum([len(trips[k]) for k in range(0, t)])
                ti1 = i1 + ti
                ti2 = ti1 + len(trips[t])

                match moving_directions[ti1]:
                    case 1:
                        index1, index2 = 0, -1
                    case 2:
                        index1, index2 = -1, 0
                    case _:
                        raise ValueError('Unkown direction')

                # first and last rail of trip
                rail_0 = rail_sequence[ti1]
                rail_n = rail_sequence[ti2 - 1]

                if track_segment_sequence[n].kilometrage.n_lines == 2 and rail_0.direction == 0:
                    if rail_sequence[ti1 - 1].direction > 0:
                        rail0_direction = rail_sequence[ti1 - 1].direction
                    elif rail_sequence[ti2].direction > 0:
                        rail0_direction = rail_sequence[ti2].direction
                    else:
                        rail0_direction = 1
                else:
                    rail0_direction = rail_0.direction
                if track_segment_sequence[n].kilometrage.n_lines == 2 and rail_n.direction == 0:
                    if rail_sequence[ti1 - 1].direction > 0:
                        railn_direction = rail_sequence[ti1 - 1].direction
                    elif rail_sequence[ti2].direction > 0:
                        railn_direction = rail_sequence[ti2].direction
                    else:
                        railn_direction = 1
                else:
                    railn_direction = rail_n.direction

                # compute km from and km to
                if n > 0 and t == 0:
                    km_from = track_segment_sequence[n].km(point=Point(rail_0.coords[index1]), direction=rail0_direction)  # type: ignore
                elif n == 0:
                    km_from = track_segment_sequence[n].km(point=gnss_coords_utm[ti1], direction=rail0_direction)  # type: ignore
                else:
                    km_from = track_segment_sequence[n].km(point=gnss_coords_utm[ti1 - 1], direction=rail0_direction)  # type: ignore
                if n < len(track_segment_indices) - 2 and t == len(trips) - 1:
                    km_to = track_segment_sequence[n].km(point=Point(rail_n.coords[index2]), direction=railn_direction)  # type: ignore
                else:
                    km_to = track_segment_sequence[n].km(point=gnss_coords_utm[ti2 - 1], direction=railn_direction)  # type: ignore

                # compute inlcine profile

                if abs(km_from - km_to) < 1e-3:
                    if len(track_segment_sequence[n].lines_1) == 1:
                        km_from = track_segment_sequence[n].km_from
                        km_to = track_segment_sequence[n].km_to

                kms_inclines, inclines = track_segment_sequence[n].incline_profile(km_from=km_from, km_to=km_to)

                trips_kms += kms_inclines
                trips_inclines += inclines
                track_segment_names += len(kms_inclines) * [track_segment_sequence[n].name]

            if len(trips_kms) > 0:
                kms_inclines_list.append(trips_kms)
                inclines_list.append(trips_inclines)
                track_segment_names_list.append(track_segment_names)

        # inclines at running km (transform kilometrage kms to running kms)
        running_kms_list = []
        for n, (kms_inclines, inclines) in enumerate(zip(kms_inclines_list, inclines_list)):
            # remove duplicates
            diff = np.diff(kms_inclines)
            index_double = np.arange(len(kms_inclines))[np.append(diff, 1) == 0]
            kms_inclines = np.delete(kms_inclines, index_double)
            inclines = np.delete(inclines, index_double)

            # moving direction
            if kms_inclines[1] - kms_inclines[0] > 0:
                direction = 1
            elif kms_inclines[1] - kms_inclines[0] < 0:
                direction = 2
            else:
                raise ValueError('Could not determine direction.')

            # transform kilometrage to points
            km_points = [track_segment_sequence[n].point(km=km_, direction=direction) for km_ in kms_inclines]

            # project points on path which yields running distances (based on path lentgth)
            if n == 0:
                running_kms = [0.0]
                distance_since_km = 0.0
            else:
                distance_since_km = abs(path.project(km_points[0]) - path.project(last_km_point)) / 1000
                running_kms = [running_kms_list[-1][-1] + distance_since_km]

            distances_along_path = path.project(km_points)
            for d1, d2 in zip(distances_along_path[1:], distances_along_path[:-1]):
                d = abs(d1 - d2) / 1000
                running_kms.append(running_kms[-1] + d)

            running_kms = np.round(running_kms, 3)
            running_kms_list.append(running_kms)

            last_km_point = km_points[-1]

        # if inclines are missing for a segment, check if
        # last available incline and next available incline are the same.
        # If yes, set nan values to the same value.
        inclines = np.concatenate((inclines_list))
        index = np.where(np.isnan(inclines))[0]
        if len(index) > 0:
            split_indices = np.where(np.diff(index) > 1)[0] + 1
            if len(split_indices) > 0:
                index_list = np.split(index, split_indices)
            else:
                index_list = [index]
            for index in index_list:
                if index[0] > 0:
                    inc1 = inclines[index[0] - 1]
                else:
                    inc1 = None
                if index[-1] < len(inclines) - 1:
                    inc2 = inclines[index[-1] + 1]
                else:
                    inc2 = None
                if inc1 == inc2:
                    inclines[index] = inc1

        # incline profile
        incline_profile = {
            'kms': np.concatenate((running_kms_list)).tolist(),
            'inclines': inclines.tolist(),
            'track_segment_names': np.concatenate((track_segment_names_list)).tolist(),
        }

        return incline_profile

    @staticmethod
    def split_into_subgroups_by_nan(
        kms: list[float], inclines: list[float], track_segment_names: list[str]
    ) -> Tuple[list[list[float]], list[list[float]], list[list[str]]]:
        """Splits kilometer values and inclines from incline profile into subgroups based on nan-values in inclines.

        Parameters
        ----------
        kms : list[float]
            The list of kilometer values.
        inclines : list[float]
            The list of inclines.
        track_segment_names : list[str]
            The list of track segment names.

        Returns
        -------
        Tuple[list[list[float]], list[list[float]], list[list[str]]]
            A tuple containing three lists: kms_splitted, inc_splitted, and tsn_splitted.
            Each list contains subgroups of kilometer values, inclines, and track segment names, respectively.
        """

        kms_splitted = []
        inc_splitted = []
        tsn_splitted = []
        sub_list_kms = []
        sub_list_inc = []
        sub_list_tsn = []
        for i in range(0, len(kms)):
            if i == 0:
                sub_list_kms.append(kms[i])
                sub_list_inc.append(inclines[i])
                sub_list_tsn.append(track_segment_names[i])

            elif (not np.isnan(inclines[i - 1])) and np.isnan(inclines[i]):  # break
                sub_list_kms.append(kms[i])
                sub_list_inc.append(inclines[i])
                sub_list_tsn.append(track_segment_names[i])
                kms_splitted.append(sub_list_kms)
                inc_splitted.append(sub_list_inc)
                tsn_splitted.append(sub_list_tsn)
                sub_list_kms = []
                sub_list_inc = []
                sub_list_tsn = []

            elif (np.isnan(inclines[i - 1])) and not np.isnan(inclines[i]):  # break
                kms_splitted.append(sub_list_kms)
                inc_splitted.append(sub_list_inc)
                tsn_splitted.append(sub_list_tsn)
                sub_list_kms = []
                sub_list_inc = []
                sub_list_tsn = []
                sub_list_kms.append(kms[i])
                sub_list_inc.append(inclines[i])
                sub_list_tsn.append(track_segment_names[i])

            else:
                sub_list_kms.append(kms[i])
                sub_list_inc.append(inclines[i])
                sub_list_tsn.append(track_segment_names[i])

            if i == len(kms) - 1:
                kms_splitted.append(sub_list_kms)
                inc_splitted.append(sub_list_inc)
                tsn_splitted.append(sub_list_tsn)

        return kms_splitted, inc_splitted, tsn_splitted

    def compute_height_profile(self, incline_profile: dict, track_segment_sequence: list[TrackSegment]) -> dict:
        """TODO"""

        kms = incline_profile['kms']
        inclines = incline_profile['inclines']
        track_segment_names = incline_profile['track_segment_names']

        # split incline profile into groups by nan values
        kms_groups, inc_groups, tsn_groups = self.split_into_subgroups_by_nan(
            kms=kms, inclines=inclines, track_segment_names=track_segment_names
        )

        # compute incline profile for each group
        height_groups = []
        searched = []
        for n, (km_group, inc_group, tsn_group) in enumerate(zip(kms_groups, inc_groups, tsn_groups)):
            # if nan group, append and continue to next group
            if np.isnan(inc_group[0]):
                height_groups.append(inc_group)
                continue

            # relative height profile
            h = [0.0]
            for i in range(1, len(km_group)):
                height_gain = (km_group[i] - km_group[i - 1]) * inc_group[i - 1]
                total_height = h[-1] + height_gain
                h.append(total_height)
            h = np.array(h)

            # find track segments for km group
            track_segment_list: list[TrackSegment] = []
            for track_segment in track_segment_sequence:
                if track_segment in track_segment_list:
                    continue

                for track_segment_name in tsn_group:
                    if track_segment.name == track_segment_name:
                        track_segment_list.append(track_segment)

            # absolute height profile
            max_height_from_incline = np.max(h)
            max_height_at_km = km_group[np.argmax(h)]
            max_height_from_isr = -1.0

            warning = False
            if n == 0:
                track_segment_list = track_segment_list[1:]
            if n == len(kms_groups) - 1:
                track_segment_list = track_segment_list[:-1]
            for track_seg in track_segment_list:
                if track_seg in searched:
                    continue
                inf_max_height = track_seg.properties['INF_HOECHSTHOEHE']
                if isinstance(inf_max_height, float):
                    if inf_max_height > max_height_from_isr:
                        max_height_from_isr = inf_max_height
                    searched.append(track_seg.name)
                else:
                    warning = True
            if max_height_from_isr == -1.0:
                max_height_from_isr = 0
                logger.debug(
                    'Max height information is missing completely in ISR. Height profile is (or parts of it are) affected by a constant offset.'
                )
            elif warning == True:
                logger.debug(
                    'Max height information partly missing in ISR. Height profile (or parts of it) may be affected by a constant offset.'
                )

            diff = max_height_from_isr - max_height_from_incline
            h_isr = h + diff

            height_groups.append(h_isr)

        # interpolate nan values for height profile
        # derive values for nan inclines from interpolated height profile
        height_groups: list[list[float]]
        kms_groups: list[list[float]]
        for i, (h_group, km_group, inc_group) in enumerate(zip(height_groups, kms_groups, inc_groups)):
            if i == 0:
                continue
            if i == len(height_groups) - 1:
                continue

            # interpolate
            if np.isnan(h_group[0]):
                h_1 = height_groups[i - 1][-1]
                h_2 = height_groups[i + 1][0]
                km_1 = kms_groups[i - 1][-1]
                km_2 = kms_groups[i + 1][0]

                h_diff = h_2 - h_1
                km_diff = km_2 - km_1
                m = h_diff / (km_diff)

                new_height_group = [h_1 + m * (km - km_1) for km in km_group]
                height_groups[i] = new_height_group

                inc_groups[i] = len(inc_groups[i]) * [m]
                inc_groups[i - 1][-1] = m

        # height and incline profile
        height_and_incline_profiles = {
            'inclines': np.concatenate((inc_groups)),
            'height': np.concatenate((height_groups)),
            'kms': incline_profile['kms'],
        }

        return height_and_incline_profiles

    def velocity_from_gnss(
        self, kms_running: list[float], gnss: GNSSSeries, incline_profile: dict, correct_velocity: bool = True
    ) -> list[float]:
        """Computes the height profile given the incline profile and track segment sequence.

        Parameters
        ----------
        incline_profile : dict
            The incline profile dictionary with keys 'kms', 'inclines', and 'track_segment_names'.
        track_segment_sequence : list[TrackSegment]
            The sequence of track segments.

        Returns
        -------
        dict
            A dictionary containing the height and incline profiles with keys 'inclines', 'height', and 'kms'.
        """

        inclines = incline_profile['inclines']
        kms = incline_profile['kms']

        # velocity_over_path
        dt_avg = np.mean(np.diff(gnss.time_s))
        velocity_over_path_kmh = []
        S = kms_running * 1000  # positions in meter
        for i in range(0, len(gnss)):
            if i == 0:
                # Use forward differences for first point
                ds = S[i + 1] - S[i]
                dt = gnss.time_s[i + 1] - gnss.time_s[i]
            elif i == len(S) - 1:
                # Use backward differences for last point
                ds = S[i] - S[i - 1]
                dt = gnss.time_s[i] - gnss.time_s[i - 1]
            else:
                # Use central differences for intermediate points
                ds = S[i + 1] - S[i - 1]
                dt = gnss.time_s[i + 1] - gnss.time_s[i - 1]

            if 2 * dt < dt_avg:  # dt can't be much smaller than average dt
                dt = dt_avg

            if correct_velocity == True:
                # get incline
                next_smaller_idx = bisect.bisect(kms, kms_running[i])
                incline = inclines[next_smaller_idx - 1]

                # correct velocity with incline angle
                alpha = np.arctan(incline / 1000)
                v = 3.6 * ds / dt / np.cos(alpha)  # km / h

            else:
                v = 3.6 * ds / dt

            velocity_over_path_kmh.append(v)

        # moving average with window length 5
        # first and last values are unfiltered
        velocity_filtered_kmh = pd.Series(velocity_over_path_kmh).rolling(5, center=True).mean().tolist()
        velocity_filtered_kmh[0] = velocity_over_path_kmh[0]
        velocity_filtered_kmh[1] = np.mean(velocity_over_path_kmh[:3])
        velocity_filtered_kmh[-2] = np.mean(velocity_over_path_kmh[-3:])
        velocity_filtered_kmh[-1] = velocity_over_path_kmh[-1]

        return velocity_filtered_kmh

    def acceleration_from_gnss_and_velocity(self, gnss: GNSSSeries, velocity_ms: list[float]) -> list[float]:
        """Computes acceleration from GNSS data and velocity.

        Parameters
        ----------
        gnss : GNSSSeries
            GNSS data series.
        velocity_ms : list[float]
            Velocity in meters per second.

        Returns
        -------
        list[float]
            Acceleration in meters per second squared.
        """

        # acceleration_over_path
        dt_avg = np.mean(np.diff(gnss.time_s))
        acceleration_over_path_ms2 = []
        for i in range(0, len(gnss)):
            if i == 0:
                # Use forward differences for first point
                dv = velocity_ms[i + 1] - velocity_ms[i]
                dt = gnss.time_s[i + 1] - gnss.time_s[i]
            elif i == len(velocity_ms) - 1:
                # Use backward differences for last point
                dv = velocity_ms[i] - velocity_ms[i - 1]
                dt = gnss.time_s[i] - gnss.time_s[i - 1]
            else:
                # Use central differences for intermediate points
                dv = velocity_ms[i + 1] - velocity_ms[i - 1]
                dt = gnss.time_s[i + 1] - gnss.time_s[i - 1]

            if 2 * dt < dt_avg:  # dt can't be much smaller than average dt
                dt = dt_avg

            acceleration_over_path_ms2.append(dv / dt)

        return acceleration_over_path_ms2

    @staticmethod
    def add_gps_series_to_map(m: folium.Map, gps: np.ndarray, df: pd.DataFrame):
        """TODO"""

        # coordinate sequence
        coordinate_sequence = [(p[0], p[1]) for p in gps]

        # extract datetime column
        datetime = df['time'].to_numpy()

        # popup for PolyLine
        # length_km = df['km_running'].to_list()[-1]                      # length of measurement ride in km
        # csv_index_start = df_subset.iloc[0].name + self.skiprows + 1     # first row index of subset in original csv file
        # csv_index_end = df_subset.iloc[-1].name + self.skiprows + 1      # last row index of subset in original csv file

        series = pd.Series(
            data=[
                f'{datetime[-1]} - {datetime[0]}',
                f'{(datetime[-1] - datetime[0]).item() / 6e10}',
                f'{len(coordinate_sequence)}',
                #           f'{csv_index_start} - {csv_index_end}',
                #            f'{length_km:.3f}',
            ],
            index=['Time', 'Duration [min]', 'Coordinates'],
            name=f'GNSS Messdaten',
        )
        df_line = series.to_frame()
        html_string = df_line.to_html(classes="table table-striped table-hover table-condensed table-responsive")
        popup = folium.Popup(html_string, max_width=500)

        scatter_group = folium.FeatureGroup(f"GNSS Messdaten").add_to(m)

        # add PolyLine
        folium.PolyLine(
            locations=coordinate_sequence,
            popup=popup,
            tooltip=f'GNSS Messdaten',
            color='black',
            weight=4,
            opacity=0.7,
            smooth_factor=0,
        ).add_to(scatter_group)

        cluster_list: list[pd.DataFrame] = []  # list for collecting rows with same location (lat, lon)
        for i, coord in enumerate(coordinate_sequence):
            # get measurement values for data point as table (will be shown as popup when clicking a point)
            row_df = df.iloc[i].to_frame(name=f'Point {df.iloc[i].name + 1}').iloc[:-5]

            if (i < len(coordinate_sequence) - 1) and (coord == coordinate_sequence[i + 1]):
                # append row
                cluster_list.append(row_df)

            else:
                row_df = pd.concat(objs=cluster_list + [row_df], axis=1)

                # popup showing data frame
                html_string = row_df.to_html(classes="table table-striped table-hover table-condensed table-responsive")
                popup = folium.Popup(html_string, max_width='600')

                # tooltip
                if len(cluster_list) > 0:
                    tooltip = f'GNSS Messdaten<br>{i+1} - {i+1+len(cluster_list)} / {len(coordinate_sequence)}'
                else:
                    tooltip = f'GNSS Messdaten<br>{i+1} / {len(coordinate_sequence)}'

                # plot measurement points as circles
                folium.Circle(
                    location=(coord[0], coord[1]),
                    radius=1,
                    color='black',
                    fill_color='lightgray',
                    opacity=0.8,
                    popup=popup,
                    tooltip=tooltip,
                ).add_to(scatter_group)

                cluster_list = []

    @staticmethod
    def add_result_to_map(
        m: folium.Map,
        df: pd.DataFrame | None,
        route,
        kms,
        gps_routed,
        distance_routed_m,
        path,
        time=None,
        name=None,
        return_map=False,
    ):
        """Adds GPS series data to a Folium map.

        Parameters
        ----------
        m : folium.Map
            The Folium map object.
        gps : np.ndarray
            The GPS coordinates.
        df : pd.DataFrame
            The DataFrame containing GPS data.

        Returns
        -------
        folium.Map | None
            If return_map is True, returns a folium.Map instance, else it returns None.
        """
        # coordinate sequence
        coordinate_sequence = [(p[0], p[1]) for p in gps_routed]

        if not isinstance(df, type(None)):
            # extract datetime column
            datetime = df['time'].to_numpy()

            # popup for PolyLine
            length_km = df['km_running'].to_list()[-1]  # length of measurement ride in km
        else:
            datetime = time

            length_km = kms[-1] - kms[0]

        if name:
            scatter_group = folium.FeatureGroup(name).add_to(m)
            color = 'green'
        else:
            scatter_group = folium.FeatureGroup(f"Ergebnis (Map Matching)").add_to(m)
            color = 'blue'
            name = 'Ergebnis (Map Matching)'

        series = pd.Series(
            data=[
                f'{datetime[-1]} - {datetime[0]}',
                f'{(datetime[-1] - datetime[0]).item() / 6e10}',
                f'{len(coordinate_sequence)}',
                #            f'{csv_index_start} - {csv_index_end}',
                f'{length_km:.3f}',
                f'{np.unique(route)}',
                f'{kms[0]} ({route[0]})',
                f'{kms[-1]} ({route[-1]})',
            ],
            index=[
                'Time',
                'Duration [min]',
                'Coordinates',
                'Ride Distance [km]',
                'Track(s)',
                'From [km (track)]',
                'To [km (track)]',
            ],
            name=name,
        )
        df_line = series.to_frame()
        html_string = df_line.to_html(classes="table table-striped table-hover table-condensed table-responsive")
        popup = folium.Popup(html_string, max_width=500)

        line_tooltip_string = (
            f'{name}<br>Track(s): {np.unique(route)}<br>From: {kms[0]} km ({route[0]}) To: {kms[-1]} km ({route[-1]})'
        )

        # add mapped measurements as line to map
        coordinate_sequence = [(p[0], p[1]) for p in gps_routed]
        path_sequence = [(p[0], p[1]) for p in path]
        folium.PolyLine(
            path_sequence,
            popup=popup,
            tooltip=line_tooltip_string,
            color=color,
            weight=3,
            opacity=0.8,
            smooth_factor=0,
        ).add_to(scatter_group)

        # plot mapped points as circles
        cluster_list: list[pd.DataFrame] = []  # list for collecting rows with same location (lat, lon)
        for i, coord in enumerate(coordinate_sequence):
            # get measurement values for data point as data frame (will be shown as popup when clicking a point)
            if not isinstance(df, type(None)):
                row_df = df.iloc[i].to_frame(name=f'Point {df.iloc[i].name + 1}')

                if (i < len(coordinate_sequence) - 1) and (coord == coordinate_sequence[i + 1]):
                    # append row
                    cluster_list.append(row_df)

                else:
                    row_df = pd.concat(objs=cluster_list + [row_df], axis=1)

                    # popup showing data frame
                    html_string = row_df.to_html(
                        classes="table table-striped table-hover table-condensed table-responsive"
                    )
                    popup = folium.Popup(html_string, max_width='600')

                    # tooltip
                    if len(cluster_list) > 0:
                        tooltip = (
                            f'Ergebnis (Map Matching)<br>{i+1} - {i+1+len(cluster_list)} / {len(coordinate_sequence)}'
                        )
                    else:
                        tooltip = f'Ergebnis (Map Matching)<br>{i+1} / {len(coordinate_sequence)}'

                    folium.Circle(
                        location=(coord[0], coord[1]),
                        radius=1,
                        color='black',
                        fill_color=color,
                        opacity=0.8,
                        popup=popup,
                        tooltip=tooltip,
                    ).add_to(scatter_group)

                    cluster_list = []
            else:
                # tooltip
                if len(cluster_list) > 0:
                    tooltip = f'{name}<br>{i+1} - {i+1+len(cluster_list)} / {len(coordinate_sequence)}'
                else:
                    tooltip = f'{name}<br>{i+1} / {len(coordinate_sequence)}'

                folium.Circle(
                    location=(coord[0], coord[1]),
                    radius=1,
                    color='black',
                    fill_color=color,
                    opacity=0.8,
                    #    popup=popup,
                    tooltip=tooltip,
                ).add_to(scatter_group)

        if return_map:
            return m

    @staticmethod
    def map_matching_results_to_map(
        df: pd.DataFrame, export_path: Path, df_path: pd.DataFrame, map_: Map | None = None, return_map: bool = False
    ) -> None | folium.Map:
        """Adds map matching results to a Folium map.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame containing map matching results.
        export_path : Path
            The path where the map will be exported.
        df_path : pd.DataFrame
            The DataFrame containing path data.
        map_ : Map | None, optional
            The map object to add ISR lines, by default None.
        return_map : bool, optional
            Whether to return the map object, by default False.

        Returns
        -------
        None | folium.Map
            None if return_map is False, otherwise the Folium map object.
        """

        # create map
        m = folium.Map(
            location=(df['lat_measured'].iloc[0], df['lon_measured'].iloc[0]),
            control_scale=True,
            tiles=None,
            zoom_start=12,
        )

        # add background layer
        folium.TileLayer(tiles='OpenStreetMap', name='OpenStreetMap Background', overlay=True).add_to(m)

        if not isinstance(map_, type(None)):
            m = MMPostprocessor.add_isr_lines_to_map(m=m, isr_map=map_, return_map=True)

        # create gps array
        latitude, longitude = df['lat_measured'].to_numpy(), df['lon_measured'].to_numpy()
        gps_array = np.column_stack((latitude, longitude))
        df_gps = df[['time', 'time_from_start_s', 'lat_measured', 'lon_measured']]

        # add measurements to map
        MMPostprocessor.add_gps_series_to_map(m, gps_array, df_gps)

        # add matched results to map
        latitude, longitude = df_path['lat_wgs84'].to_numpy(), df_path['lon_wgs84'].to_numpy()
        path = np.column_stack((latitude, longitude))
        latitude, longitude = df['lat_wgs84'].to_numpy(), df['lon_wgs84'].to_numpy()
        gps_array = np.column_stack((latitude, longitude))
        MMPostprocessor.add_result_to_map(
            m, df, df['track_number'].to_numpy(), df['km_db'].to_numpy(), gps_array, df['gnss_error_m'].to_numpy(), path
        )

        if return_map:
            return m
        else:
            # save map
            folium.LayerControl().add_to(m)
            m.save(export_path / f'map.html')

    @staticmethod
    def add_isr_lines_to_map(m: folium.Map, isr_map: Map, return_map: bool = False):
        """Adds ISR lines and operational points to a Folium map.

        Parameters
        ----------
        m : folium.Map
            The Folium map object.
        isr_map : Map
            The ISR map containing rails and operational points.
        return_map : bool, optional
            Whether to return the map object, by default False.

        Returns
        -------
        None | folium.Map
            None if return_map is False, otherwise the Folium map object.
        """
        colors = [
            'red',
            #           'blue',
            'orange',
            #            'beige',
            'purple',
            'green',
            'darkred',
            #            'darkblue',
            #            'pink',
            'lightred',
            'lightblue',
            'darkgreen',
            'lightgreen',
            'cadetblue',
            'darkpurple',
            #            'lightgray',
            #            'gray',
            #            'black'
        ]
        color_cycle = cycle(colors)

        # sort rails by track number
        rails = isr_map['rails']
        tracks = [int(rail.track) for rail in rails]
        index = np.argsort(tracks).tolist()
        rails: list[Rail] = [rails[i] for i in index]  # type: ignore

        scatter_group = folium.FeatureGroup(f'ISR Strecken').add_to(m)
        # add rails to map in groups based on track number
        for i, rail in enumerate(rails):
            if i == 0 or rails[i - 1].track != rails[i].track:
                color = next(color_cycle)

            coordinate_sequence = [(p.x, p.y) for p in rail.coords_wgs]
            folium.PolyLine(
                coordinate_sequence,
                popup=f'Strecke {rail.track}',
                tooltip=f'Strecke {rail.track}',
                color=color,
                weight=12,
                opacity=0.6,
                smooth_factor=0,
            ).add_to(scatter_group)

        # add betriebsstellen to map
        scatter_group_bs = folium.FeatureGroup('Betriebsstellen (ISR)').add_to(m)

        for bs in isr_map['operational_points']:
            folium.Circle(
                location=(bs.coords_wgs84[0], bs.coords_wgs84[1]),
                radius=8,
                color='green',
                fill_color='green',
                opacity=1,
                popup=f'{bs.name}',
                tooltip=f'{bs.name}',
            ).add_to(scatter_group_bs)

        if return_map == True:
            return m



from pathlib import Path
from typing import Literal, Union, Tuple
import numpy as np
import pandas as pd
from isr_matcher.geometry.gnss_series import GNSSSeries
from datetime import datetime


class GNSSDataImport:

    """
    Class that manages the import of GPS data.

    Attributes
    ----------
    TODO

    """

    def __init__(self):
        self.column_names_obligatory = [['time_utc', 'time_s'], 'latitude', 'longitude']
        self.column_names_optional = ['altitude_m', 'velocity_ms', 'acceleration_ms2']

    def read_csv(
        self,
        csv_path: Path | str,
        column_name_dict: dict | None,
        delimiter: str = ',',
        decimal: str = '.',
        na_values: list[str] | None = None,
        skiprows: int = 0,
    ):
        """
        "time_utc": "time_utc",
        "time_s": "measured_time_1_in_s",
        "latitude": "latitude",
        "longitude": "longitude",
        "altitude_m": None,
        "velocity_ms": "speed_ms",
        "acceleration_ms2": None,
        """

        # assert input
        if not isinstance(column_name_dict, type(None)):
            if column_name_dict['time_s'] and column_name_dict['time_utc']:
                # both times
                parse_dates = [column_name_dict['time_utc']]

            elif column_name_dict['time_s']:
                # no utc time
                parse_dates = False

            elif column_name_dict['time_utc']:
                # compute time in seconds
                parse_dates = [column_name_dict['time_utc']]
                time_s = None
            else:
                raise ValueError(
                    "Both 'time_utc' and 'time_s' in column_name_dict are None. At least one must be specified: The csv file must contain at least one time column. Please set the corresponding column name in column_name_dict."
                )
        else:
            parse_dates = ['time_utc']

        try:
            # read
            df = pd.read_csv(
                csv_path,
                delimiter=delimiter,
                decimal=decimal,
                parse_dates=parse_dates,
                na_values=na_values,
                skiprows=skiprows,
            )

        except:
            # read
            df = pd.read_csv(
                csv_path,
                delimiter=delimiter,
                decimal=decimal,
                parse_dates=False,
                na_values=na_values,
                skiprows=skiprows,
            )

        for column in self.column_names_obligatory:
            if isinstance(column, list):
                assert (
                    column[0] in df.columns or column[1] in df.columns
                ), 'One time column is missing in csv file or has wrong name. Please check the file or the column_name_dict parameter.'
            else:
                assert (
                    column in df.columns
                ), f'Column {column} is missing in csv file or incorrectly mapped with parameter column_name_dict.'

        if not isinstance(column_name_dict, type(None)):
            # parse
            column_names_dict = {
                val: key for (key, val) in zip(column_name_dict.keys(), column_name_dict.values()) if val
            }
            column_names = [val for val in column_name_dict.values() if val]
            column_names_dict = {
                val: key for (key, val) in zip(column_name_dict.keys(), column_name_dict.values()) if val
            }

            # rename to standardized column names
            df = df[column_names]
            df = df.rename(columns=column_names_dict)

        # average over identical positions
        aggregate = []
        for column in df.columns:
            aggregate.append((column, np.mean))
        df = df.groupby(['latitude', 'longitude'], sort=False).agg('mean').reset_index()

        # extract values
        time_utc = df['time_utc'].to_numpy() if 'time_utc' in df.columns else None
        time_s = df['time_s'].to_numpy() if 'time_s' in df.columns else None
        altitude_m = df['altitude_m'].to_numpy() if 'altitude_m' in df.columns else None
        velocity_ms = df['velocity_ms'].to_numpy() if 'velocity_ms' in df.columns else None
        acceleration_ms2 = df['acceleration_ms2'].to_numpy() if 'acceleration_ms2' in df.columns else None

        # return instance
        return GNSSSeries(
            time_utc=time_utc,
            time_s=time_s,
            latitude=df['latitude'].to_numpy(),
            longitude=df['longitude'].to_numpy(),
            altitude_m=altitude_m,
            velocity_ms=velocity_ms,
            acceleration_ms2=acceleration_ms2,
        )



from __future__ import annotations
import numpy as np
from numpy import newaxis as na
from shapely import Point
from shapely.ops import nearest_points
from isr_matcher.geometry.map import Map
from isr_matcher.geometry.gnss_series import GNSSSeries
from isr_matcher.geometry.rail import Rail
from isr_matcher.ops._functions import angle_between
from isr_matcher._constants._filters import ISR_EXCEPTIONAL_STATION_NAMES
from isr_matcher._constants.logging import setup_logger
import logging

# Create logger for the current module
setup_logger()
logger = logging.getLogger(__name__)


class HMM:
    """Class representing a Hidden Markov Model (HMM).

    Attributes
    ----------
    N : int
        Number of states.
    M : int
        Number of observations.
    A_list : list[np.ndarray]
        List of state transition probability matrices.
    B_log : np.ndarray
        Emission probability matrix.
    C_log : np.ndarray
        Initial distribution.
    """

    def __init__(self, A_list: list[np.ndarray], B_log: np.ndarray, C_log: np.ndarray):
        """Initializes the Hidden Markov Model with the given parameters.

        Parameters
        ----------
        A_list : list[np.ndarray]
            List of state transition probability matrices, each shape (N,N).
        B_log : np.ndarray
            Emission probability matrix (log), shape (N,M).
        C_log : np.ndarray
            Initial distribution (log), shape(N,).
        """
        self.N = A_list[0].shape[0]  # number of states
        self.M = B_log.shape[1]  # number of observations

        self.A_list = A_list  # list of state transition probability matrices, list of length M, dimension: (N, N)
        self.B_log = B_log  # emission probability matrix,         dimension: (N, M)
        self.C_log = C_log  # initial distribution,                dimension: (N,)

    def solve(self) -> list[int]:
        """Solves HMM for optimal state sequence using Viterbi algorithm (in log-domain).

        Original implementation from: https://www.audiolabs-erlangen.de/resources/MIR/FMP/C5/C5S3_Viterbi.html

        Returns
        -------
        S_opt: list[int]
            Optimal state sequence of length M
        """

        N = self.N  # Number of states
        M = self.M  # Number of observations
        A_list = self.A_list  # list of transition matrices (one matrix for each pair of observations)
        B_log = self.B_log  # emission probability matrix (log)
        C_log = self.C_log  # initial state distribution (log)

        # Initialize D and E matrices
        D_log = np.zeros((N, M))
        E = np.zeros((N, M - 1)).astype(np.int32)
        D_log[:, 0] = C_log + B_log[:, 0]

        # Compute D and E in a nested loop
        for t in range(1, M):
            A_log = np.log(A_list[t - 1])  # transition matrix for timestep

            for i in range(N):
                temp_sum = A_log[:, i] + D_log[:, t - 1]
                D_log[i, t] = np.max(temp_sum) + B_log[i, t]
                E[i, t - 1] = np.argmax(temp_sum)

        # Backtracking
        S_opt = np.zeros(M).astype(np.int32)
        S_opt[-1] = np.argmax(D_log[:, -1])
        for t in range(M - 2, -1, -1):
            S_opt[t] = E[int(S_opt[t + 1]), t]

        return S_opt.tolist()

    @classmethod
    def state_transition_matrix_from_map(cls, map_: Map, gnss: GNSSSeries) -> list[np.ndarray]:
        """
        This method calculates state transition probability matrices for each time step based on the given map and GNSS series.

        Parameters
        ----------
        map_ : Map
            The map containing information about rails, operational points, and track transitions.
        gnss : GNSSSeries
            The GNSS series containing coordinate observations.

        Returns
        -------
        A_list: list[np.ndarray]
            A list of state transition probability matrices for each time step, excluding the initial time step.
        """

        rails = map_['rails']  # all rails in map
        ops = map_['operational_points'] + map_['track_transitions']
        observations = gnss.coords_utm  # observations as sequence of Points
        N = len(rails)  # number of rails (states)
        M = len(observations)  # number of gnss coordinates (observations)
        A_list = []  # list of transition matrices, for each t = 1,...,M (skipping 0)
        dr = 3.1

        # for each observation
        logger.info(f'Computing transition matrices for {M} time steps...')
        for t in range(1, M):
            A = np.zeros((N, N))  # initialize state transition matrix for time step t
            obs_current = observations[t]  # observation at time step t
            obs_previous = observations[t - 1]  # observation at time step t-1

            for i in range(N):
                # for each state / rail, we compute the transition probabilites to all other rails

                rail_i = rails[i]  # current rail
                x_i_current = nearest_points(rail_i, obs_current)[0]  # current obs projected on rail
                x_i_previous = nearest_points(rail_i, obs_previous)[0]  # prev obs projected on rail
                rail_i_start_point = Point(rail_i.coords[0])  # rail start coordinates
                rail_i_end_point = Point(rail_i.coords[-1])  # rail end coordinates

                # distance along rails for projections
                d_x_i_current = rail_i.project(x_i_current, normalized=True)  # distance along rail for current obs
                d_x_i_previous = rail_i.project(x_i_previous, normalized=True)  # distance along rail for prev obs

                # driving direction on rail i (source) based on observations
                if d_x_i_current - d_x_i_previous > 0:
                    driving_direction_i = 1  # driving direction 1 : direction of increasing km
                elif d_x_i_current - d_x_i_previous < 0:
                    driving_direction_i = 2  # driving direction 2 : direction of decreasing km
                else:
                    driving_direction_i = 0  # invalid ( direction not determinable )

                match driving_direction_i:
                    # set end point according to driving direction
                    case 0:
                        end_point = None
                    case 1:
                        end_point = rail_i_end_point
                    case 2:
                        end_point = rail_i_start_point
                    case _:
                        raise NotImplementedError(f'Invalid value for driving direction: {driving_direction_i}')

                if end_point and end_point.distance(x_i_current) < 1:  # transition between rails (current rail ends)
                    for j in range(N):
                        if i == j:
                            continue  # transition between rails, so no transition on same rail
                        if rail_i.distance(rails[j]) > 100.0:
                            continue  # this sets a threshold for the maximum length of a connecting rail!
                        rail_j = rails[j]

                        # allow transition if current rail and target rail are close to the same BS
                        # -> Missing junctions can cause HMM breaks, in stations often many junctions are missing in ISR
                        if rail_i.track != rail_j.track:
                            for op in ops:
                                if op.distance(rail_i) < 20 and op.distance(rail_j) < 20:
                                    connecting_rail = True
                                    reduce_factor = 1e-6
                                    break
                            else:
                                connecting_rail = False
                                reduce_factor = 1.0
                        else:
                            connecting_rail = False
                            reduce_factor = 1.0

                        rail_j_start_point = Point(rail_j.coords[0])  # target rail start point
                        rail_j_end_point = Point(rail_j.coords[-1])  # target rail end point
                        x_j_current = nearest_points(rail_j, obs_current)[0]  # projection on target rail

                        # no transitions to a rail that ends parallel to current rail
                        if driving_direction_i == 1 and (x_j_current == rail_j_end_point):
                            continue
                        elif driving_direction_i == 2 and (x_j_current == rail_j_start_point):
                            continue
                        elif rail_i_end_point == rail_j_end_point:
                            continue

                        # continue if end points are further than dr apart and there is no connecting rail
                        # (reason: if observations are sparse, rails may be skipped)
                        distance = min(end_point.distance([rail_j_start_point, rail_j_end_point]))

                        if connecting_rail == False:
                            for k in range(len(rails)):
                                if i == k or j == k:
                                    continue  # skip identical rail
                                if max(rails[k].distance([rails[i], rails[j]])) > dr:
                                    continue  # if min. distances is greater than dr, k cannot be connecting rail

                                rail_k_start_point = Point(rails[k].coords[0])  # conn. rail start point
                                rail_k_end_point = Point(rails[k].coords[-1])  # conn. rail end point

                                # distance between end points of rail i and rail k < dr
                                distance_ik = end_point.distance([rail_k_start_point, rail_k_end_point])
                                min_index = np.argmin(distance_ik)
                                connection_ik = distance_ik[min_index] < dr
                                if connection_ik == False:
                                    continue

                                # distance between end points of rail k and rail j < dr
                                other_index = int(min_index) - 1  # 0 -> -1, 1 -> 0
                                distance_kj = [rail_k_start_point, rail_k_end_point][other_index].distance(
                                    [rail_j_start_point, rail_j_end_point]
                                )
                                connection_kj = min(distance_kj) < dr
                                if connection_kj == False:
                                    continue

                                # direction vector rail i
                                end_point_arr = np.array([end_point.x, end_point.y])
                                if end_point == rail_i_start_point:
                                    prev_point_arr = np.array([rail_i.coords[1][0], rail_i.coords[1][1]])
                                elif end_point == rail_i_end_point:
                                    prev_point_arr = np.array([rail_i.coords[-2][0], rail_i.coords[-2][1]])
                                else:
                                    raise NotImplementedError()
                                rail_i_direction_vec = end_point_arr - prev_point_arr

                                # direction vector rail k
                                if min_index == 0:
                                    prev_point_arr = np.array([rails[k].coords[1][0], rails[k].coords[1][1]])
                                elif min_index == 1:
                                    prev_point_arr = np.array([rails[k].coords[-2][0], rails[k].coords[-2][1]])
                                else:
                                    raise NotImplementedError()
                                rail_k_direction_vec = prev_point_arr - end_point_arr

                                # angle rail i rail k
                                angle_ik = angle_between(rail_i_direction_vec, rail_k_direction_vec)
                                if angle_ik > 90:
                                    continue  # no connecting rail that would need vehicle to do a u-turn

                                connecting_rail = True
                                break

                            else:
                                connecting_rail = False

                        # continue if end points are further than dr apart and there is no connecting rail
                        if distance > dr and connecting_rail == False:
                            continue

                        # previous code in loop is only to exclude rails to which transitioning does not make sense
                        # at this point, we specify the transition probability between rail i and j (at current time step)
                        # we consider three parameters

                        #   - track number: higher prob. for transitioning to a rail of the same track
                        #                   or at track transitions ('streckenwechsel')

                        transitions = [
                            str(rail_i.track) in tt.name
                            for tt in map_['track_transitions']
                            if tt.distance(rail_j_start_point) < 5
                        ]
                        p_track = 0.8 if rail_i.track == rail_j.track or any(transitions) else 0.2

                        #   - driving direction: higher prob. for transitioning to a rail with same direction (direction 1->2 and 2->1 unlikely)

                        p_direction = 0.9 if driving_direction_i + rail_j.direction != 3 else 0.1

                        #  - distance: higher prob. to transition to closer rails and highest prob. to connected rails or free ends,

                        if all(
                            [
                                min(rail.distance([rail_j_start_point, rail_j_end_point])) > 1
                                for rail in rails
                                if rail != rail_j
                            ]
                        ):
                            distance = 0  # set distance=0 for free ends
                        lambd = 0.5  # parameter
                        p_distance = np.exp(-lambd * distance)  # higher prob. for lower distance

                        # total probability
                        A[i, j] = p_track * p_direction * p_distance * reduce_factor

                else:  # no transition / self-transition
                    # self-transitions are much more common than transitions to other rails
                    # this is especially true for high sample rates
                    # therefore, the probabilites of self-transitions have much more influence
                    # than the probabilities of rail-transitions, which are only punctual
                    # thus, we encode direction information also in the self-transition probabilities,
                    # setting a higher prob. for rail whose direction matches the driving direction

                    ts_name = rail_i.track_segment_name
                    if ts_name.count(' - ') > 1:
                        a, b, c = ts_name.split(' - ')
                        if a + ' - ' + b in ISR_EXCEPTIONAL_STATION_NAMES:
                            from_name = a + ' - ' + b
                            to_name = c
                        else:  # b + ' - ' + c in ISR_EXCEPTIONAL_STATION_NAMES:
                            from_name = a
                            to_name = b + ' - ' + c
                    else:
                        from_name, to_name = ts_name.split(' - ')
                    p_same = (
                        0.5 if 'Hbf' in from_name and 'Hbf' in to_name else 0.9
                    )  # at main stations, ignore rail direction

                    # only allow self transitions
                    if driving_direction_i + rails[i].direction != 3:
                        A[i, i] = p_same  # higher prob. if driving direction matches rail direction
                    else:
                        A[i, i] = 1 - p_same  # lower prob. otherwise
                    # -> NO Normalization (rows)!

            A_list.append(A)  # append to list
        logger.info(f'Transition matrices computed.')
        return A_list

    @staticmethod
    def emission_probabilities(rail: Rail, measurements: list[Point], sigma: float) -> np.ndarray:
        """Computes emission probabilities for the given rail, observations, and standard deviation of GNSS noise.

        Parameters
        ----------
        rail : Rail
            The rail for which emission probabilities are computed.
        measurements : list[Point]
            The sequence of observations.
        sigma : float
            The standard deviation of GNSS noise.

        Returns
        -------
        np.ndarray
            The emission probabilities for the given rail and observations.
        """
        exponent = -0.5 * (rail.distance(measurements) / sigma) ** 2
        return np.exp(exponent) / (np.sqrt(2 * np.pi) * sigma)

    @staticmethod
    def log_emission_probabilities(rail: Rail, measurements: list[Point], sigma: float) -> np.ndarray:
        """Computes emission probabilities for state 'rail', observations 'measurements' and standard deviation of GNSS noise 'sigma' directly in log space.

        Gaussian distribution in log space: log( 1 / (sqrt(2*pi) * sigma) ) - 0.5 (distance / sigma)^2
        """
        return np.log(1 / (np.sqrt(2 * np.pi) * sigma)) - 0.5 * (rail.distance(measurements) / sigma) ** 2

    @classmethod
    def emission_probability_matrix_from_map_and_gnss(cls, map_: Map, gnss: GNSSSeries) -> np.ndarray:
        """Computes emission probability matrix in log-domain from Map 'map_' and GNSS measurements 'gnss'.

        Parameters
        ----------
        rail : Rail
            The rail for which emission probabilities are computed.
        measurements : list[Point]
            The sequence of observations.
        sigma : float
            The standard deviation of GNSS noise.

        Returns
        -------
        B_log: np.ndarray
            The emission probabilities for the given rail and observations in log space.
        """
        rails = map_['rails']  # all rails in map
        N = len(rails)  # number of rails (states)
        M = len(gnss)  # number of observations
        B_log = np.zeros((N, M))  # initialize state transition matrix

        # compute and set emission probabilites
        logger.info(f'Computing emission probabilites for {N} rails...')
        for i in range(N):
            B_log[i, :] = HMM.log_emission_probabilities(rail=rails[i], measurements=gnss.coords_utm, sigma=gnss.sigma)  # type: ignore (gnss.sigma is never None at this stage)
            B_log[i, :] = np.where(
                rails[i].distance(gnss.coords_utm) > 500.0, -np.inf, B_log[i, :]
            )  # set emission prob. to 0 for distant rails
        logger.info(f'Emission probabilites computed.')
        return B_log

    @classmethod
    def from_map_and_gnss(cls, map_: Map, gnss: GNSSSeries) -> HMM:
        """Constructs an HMM from a map instance 'map_' and GNSS measurements 'gnss'.

        Parameters
        ----------
        map_: Map
            Instance of class Map (TypedDict with keys 'track_segments', 'operational_points' and 'rails').
        gnss: GNSSSeries
            Instance of class GNSSSeries, representing the measured GNSS parameters (lat, lon, utm coordinates, velocity, etc.).

        Returns
        -------
        hmm: HMM
            A new instance of HMM with parameters (A, B, C) computed from 'map_' and GNSS measurements 'gnss'.
        """

        # compute HMM parameter
        A_list = cls.state_transition_matrix_from_map(map_=map_, gnss=gnss)
        B_log = cls.emission_probability_matrix_from_map_and_gnss(map_=map_, gnss=gnss)
        C_log = B_log[:, 0]

        return cls(A_list, B_log, C_log)
